//----------------------------------------------------------------------------
//	SOURCE:	MapArea.c
//	AUTHOR:	Ian Taylor
//	NOTES:	Data source file of Map2 demo application
//----------------------------------------------------------------------------

#include "Map.h"
#include "MapCodes.h"

// Name data
LPCTSTR region_names[Code_Final];

local void DecompressCountryNames(void)
{
	// Decompress the countries
#define COUNTRY(a,b,d)	b "\0"
#define REGION(a,b,c)	c "\0"
	LPCTSTR s =
#include "MapTable.h"
;
	int i;
	for (i = 1; i < Code_Final; i++)
	{
		region_names[i] = s;
		while (*(s++) != '\0');
	}
	return;
}

int RegionToCountry(int region)
{
	// Work out the country (and region) name of this index
// Region-to-country mapping
#define COUNTRY(a,b,d)
#define REGION(a,b,c)	Code_##a,
local const BYTE region_country[Code_RegionCount] = {
#include "MapTable.h"
};
	ASSERT((region >= 0) && (region < dimof(region_names)));
	if (region < Code_RegionFirst)
		return region;
	return region_country[region - Code_RegionFirst];
}

// Encoding/decoding floats in bytes
#define ENFLOAT3(x)		(x & 0xFF), ((x >> 8) & 0xFF), ((x >> 16) & 0xFF)
#define ENFLOAT1(x)		ENFLOAT3((DWORD)(x * 0xFFFFFF))
#define ENFLOAT(x,l,u)	ENFLOAT1((x + l) / (u - l))
#define DEFLOAT3(x)		((DWORD)x[0] + ((DWORD)x[1] << 8) + ((DWORD)x[2] << 16))
#define DEFLOAT1(x)		((double)DEFLOAT3(x) / 0xFFFFFF)
#define DEFLOAT(x,l,u)	(float)(DEFLOAT1(x) * (u - l) + l)

// Vector data
#define AREA(name)		Code_##name,
#define LOOP(x,y)		0xF6, ENFLOAT(x,-181,181), ENFLOAT(y,-90,90),
#define LINE__			0xF7,
#define LINE_(id)		0xF8,
#define LINE(id)		SEGMENT(id)
#define SEGMENT__		0xF9,
#define SEGMENT_(id)	
#define SEGMENT(id)		(((id >> 8) & 0xFF) | 0xF0), (id & 0xFF),
#define SEGMENT_END		0xFA,
#define TRIANGLE(x,y)	0xFB, ENFLOAT(x,-181,181), ENFLOAT(y,-90,90),
#define POINTS_11(a,b)	0xEE,
#define POINTS_10(a,b)	(0xE0 | (b >> 8)), (b & 0xFF),
#define POINTS_01(a,b)	(0x0E | ((a >> 8) << 4)), (a & 0xFF),
#define POINTS_00(a,b)	(((a >> 8) << 4) | (b >> 8)), (a & 0xFF), (b & 0xFF),
#define POINTS__1(a,b)	0xEE,
#define POINTS__0(a,b)	(0xE0 | (b >> 8)), (b & 0xFF),
#define POINT(x,y)		ENFLOAT(x,-181,181), ENFLOAT(y,-90,90),
#define LENGTH(d)		(BYTE)(d * 100 + 0.5),
#define TWIST(a)		(BYTE)(a / 120 * 256),

local const BYTE compressed[] =
{
#include "MapData.h"
0
};
#undef AREA

// Used in MapFlag
#undef TRIANGLE

// All the region polygons in lon/lat coords
local MAPPOINT area_point[dimof(compressed)];

// The raw instruction list
int raw_instruction[MAX_INSTRUCTIONS];
int raw_instructions;

// The raw area list
MAPAREA raw_area[MAX_AREAS];
int raw_areas;

// The raw point list
MAPPOINT raw_point[MAX_POINTS];
int raw_points;

// Transits
typedef struct _TRANSIT
{
	int in_edge;
	MAPPOINT in_point;
	int in_offset;
	int out_edge;
	MAPPOINT out_point;
	int out_offset;
}
TRANSIT;

// Remembered points and segments during decompression
typedef struct _REMEMBERED
{
	MAPPOINT point[512];
	int points;
	struct _REMSEGMENT
	{
		const MAPPOINT* first;
		const MAPPOINT* last;
	}
	segment[512];
	int segments;
}
REMEMBERED;

// Debugging only
#ifdef _DEBUG
#include <stdio.h>
#include <stdarg.h>
void DEBUG(const char* fmt, ...)
{
	static int pause = 0;
	char buffer[1024];
	va_list vargs;
	va_start(vargs, fmt);
	vsprintf(buffer, fmt, vargs);
	va_end(vargs);
	OutputDebugString(buffer);
	if (++pause >= 10)
	{
		Sleep(10);
		pause = 0;
	}
	return;
}
#define DEBUG1(params) //DEBUG params
#define DEBUG2(params) //DEBUG params
#else
#define DEBUG1(params)
#define DEBUG2(params)
#endif


local BOOL DecompressPointsEqual(const MAPPOINT* a, const MAPPOINT* b)
{
	return (a->x == b->x) && (a->y == b->y);
}

local const BYTE* DecompressTriangle(const BYTE* cmp, int first)
{
	// Decompress a triangle
	double x = area_point[first].x;
	double y = area_point[first].y;
	double d = *(cmp++) * 0.01;
	double angle = (*(cmp++) * M_PI) / 384;
	int i;
	DEBUG1((" LENGTH(%g)\r\n", d));
	for (i = first; i < (first + 3); i++)
	{
		area_point[i].x = (float)(x + d * cos(angle));
		ASSERT((area_point[i].x > -180) && (area_point[i].x < 180));
		area_point[i].y = (float)(y + d * sin(angle));
		ASSERT((area_point[i].y > 0) && (area_point[i].y < 180));
		angle -= M_PI * 2 / 3;
	}
	return cmp;
}

local const BYTE* DecompressPolyline(const BYTE* cmp, int first, int* last)
{
	// Decompress a polyline
	double x = area_point[first].x;
	double y = area_point[first].y;
	double d = *(cmp++) * 0.01;
	int n = first;
	DEBUG1((" LENGTH(%g)\r\n      ", d));
	do
	{
		double angle = *cmp * M_PI / 128;
		x += sin(angle) * d;
		ASSERT((x >= -181) && (x <= 181));
		y += cos(angle) * d;
		ASSERT((y >= 0) && (y <= 180));
		n++;
		area_point[n].x = (float)x;
		area_point[n].y = (float)y;
		DEBUG1(("%d[%g,%g],", *cmp, x, y));
		cmp++;
	}
	while (cmp[0] != (cmp[-1] ^ 0x80));
	*last = n;
	DEBUG1(("%d\r\n", *cmp));
	return cmp + 1;
}

local const BYTE* DecompressCoord(const BYTE* cmp, MAPPOINT* pt)
{
	// Decompress coordinates
	pt->x = DEFLOAT(cmp, -181, 181);
	cmp += 3;
	pt->y = DEFLOAT(cmp, 0, 180);
	cmp += 3;
	return cmp;
}

local const BYTE* DecompressPoints(const BYTE* cmp, const MAPPOINT** v1, const MAPPOINT** v2, const MAPPOINT* prev, REMEMBERED* rem)
{
	// Fetch two (possibly remembered) points
	BYTE c = *(cmp++);
	int c1 = c & 0xF0;
	int c2 = c & 0x0F;
	DEBUG1(("    POINTS(a,b)"));
	if (c1 != 0xE0)
	{
		// Fetch the remembered point
		int i1 = (c1 << 4) + *(cmp++) - 1;
		ASSERT(i1 < rem->points);
		*v1 = &rem->point[i1];
	}
	if (c2 != 0x0E)
	{
		// Fetch the remembered point
		int i2 = (c2 << 8) + *(cmp++) - 1;
		ASSERT(i2 < rem->points);
		*v2 = &rem->point[i2];
	}
	if (c1 == 0xE0)
	{
		// The first one is new or previous, so read it
		if (prev != NULL)
		{
			// It's a previous point
			*v1 = prev;
		}
		else
		{
			// It's an explicit point
			cmp = DecompressCoord(cmp, &rem->point[rem->points]);
			*v1 = &rem->point[rem->points++];
		}
		DEBUG1((" POINT(%g,%g)", (*v1)->x, (*v1)->y));
	}
	if (c2 == 0x0E)
	{
		// The second one is new, so read it
		cmp = DecompressCoord(cmp, &rem->point[rem->points]);
		*v2 = &rem->point[rem->points++];
		DEBUG1((" POINT(%g,%g)", (*v2)->x, (*v2)->y));
	}
	DEBUG1((" [%g,%g] [%g,%g]", (*v1)->x, (*v1)->y, (*v2)->x, (*v2)->y));
	return cmp;
}

local void DecompressAreas(void)
{
	// Static because it's large enough to cause a stack frame expansion
	static REMEMBERED rem;
	const MAPPOINT* v1;
	const MAPPOINT* v2;
	const BYTE* cmp = compressed;
	int pts, last, n = 0;

	// Decompress all the areas
	memset(&rem, 0, sizeof(rem));
	raw_areas = 1;
	raw_area[0].point_offset = 0;
	raw_area[1].point_offset = 0;
	do
	{
		BOOL finished = FALSE;
		const MAPPOINT* prev = NULL;

		// Fetch the next area
		DEBUG1(("AREA(%d, '%s')\r\n", raw_areas, region_names[*cmp]));
		raw_area[raw_areas].code = *(cmp++);
		do
		{
			switch (*cmp)
			{
				case 0xF6:
					// This is a loop
					ASSERT(prev == NULL);
					cmp++;

					// Fetch the start point
					cmp = DecompressCoord(cmp, &area_point[n]);
					DEBUG1(("  LOOP(%g,%g)", area_point[n].x, area_point[n].y));

					// Fetch the loop points
					cmp = DecompressPolyline(cmp, n, &last);
					area_point[last + 1] = area_point[n];
					n = last + 2;
					finished = TRUE;
					break;

				case 0xF7:
					// This is an anonymous line
					DEBUG1(("  LINE__\r\n"));
					cmp++;

					// Fetch the start and end points
					cmp = DecompressPoints(cmp, &v1, &v2, prev, &rem);
					if (prev == NULL)
						area_point[n++] = *v1;
					else
						ASSERT(DecompressPointsEqual(&area_point[n-1], v1));
					area_point[n++] = *v2;
					prev = v2;
					DEBUG1(("\r\n"));
					break;

				case 0xF8:
					// This is an remembered line
					DEBUG1(("  LINE_(%d)\r\n", rem.segments + 1));
					cmp++;

					// Fetch the start and end points
					cmp = DecompressPoints(cmp, &v1, &v2, prev, &rem);
					ASSERT(rem.segments < dimof(rem.segment));
					if (prev == NULL)
						area_point[n++] = *v1;
					else
						ASSERT(DecompressPointsEqual(&area_point[n-1], v1));
					rem.segment[rem.segments].first = &area_point[n-1];
					area_point[n] = *v2;
					prev = v2;
					rem.segment[rem.segments].last = &area_point[n++];
					rem.segments++;
					DEBUG1(("\r\n"));
					break;

				case 0xF9:
					// This is an anonymous segment
					DEBUG1(("  SEGMENT__\r\n"));
					cmp++;

					// Fetch the start and end points
					cmp = DecompressPoints(cmp, &v1, &v2, prev, &rem);

					// Store all the points
					if (prev == NULL)
						area_point[n++] = *v1;
					else
						ASSERT(DecompressPointsEqual(&area_point[n-1], v1));
					cmp = DecompressPolyline(cmp, n-1, &last);
					n = last + 1;
					area_point[n++] = *v2;
					prev = v2;
					break;

				case 0xFA:
					// This is a segment end
					DEBUG1(("  SEGMENT_END\r\n"));
					cmp++;
					finished = TRUE;
					break;

				case 0xFB:
					// This is a triangle
					ASSERT(prev == NULL);
					cmp++;

					// Fetch the start point
					cmp = DecompressCoord(cmp, &area_point[n]);
					DEBUG1(("  TRIANGLE(%g,%g)", area_point[n].x, area_point[n].y));

					// Fetch the loop points
					cmp = DecompressTriangle(cmp, n);
					area_point[n + 3] = area_point[n];
					n += 4;
					finished = TRUE;
					break;

				default:
					// Non-special marker
					if (*cmp >= 0xF0)
					{
						// This is a recalled segment
						const MAPPOINT* p;
						int id = ((int)cmp[0] << 8) + cmp[1];
						if (id & 0x0800)
						{
							// Negative
							id |= ~0x0FFF;
							DEBUG1(("  SEGMENT(%d)\r\n", id));
							id = -id;
							ASSERT((id > 0) && (id <= rem.segments));
							p = rem.segment[id-1].first;
							do
								area_point[n++] = *p;
							while (p++ != rem.segment[id-1].last);
						}
						else
						{
							// Positive
							id &= 0x0FFF;
							DEBUG1(("  SEGMENT(%d)\r\n", id));
							ASSERT((id > 0) && (id <= rem.segments));
							p = rem.segment[id-1].last;
							do
								area_point[n++] = *p;
							while (p-- != rem.segment[id-1].first);
						}
						prev = &area_point[n-1];
						cmp += 2;
					}
					else
					{
						// This is an remembered segment
						// Fetch the start and end points
						DEBUG1(("  SEGMENT_(%d)\r\n", rem.segments + 1));
						cmp = DecompressPoints(cmp, &v1, &v2, prev, &rem);
						ASSERT(rem.segments < dimof(rem.segment));
						if (prev == NULL)
							area_point[n++] = *v1;
						else
							ASSERT(DecompressPointsEqual(&area_point[n-1], v1));
						rem.segment[rem.segments].first = &area_point[n-1];

						// Fetch the intermediate points
						cmp = DecompressPolyline(cmp, n-1, &last);
						n = last + 1;
						area_point[n] = *v2;
						prev = v2;
						rem.segment[rem.segments].last = &area_point[n++];
						rem.segments++;
					}
					break;
			}
			DEBUG1(("--- raw_areas=%d n=%d\r\n", raw_areas, n));
		}
		while (!finished);

		// We've completed the polygon
		ASSERT(DecompressPointsEqual(&area_point[raw_area[raw_areas].point_offset], &area_point[n-1]));
		pts = n - raw_area[raw_areas].point_offset;
		ASSERT(pts > 3);
		raw_area[++raw_areas].point_offset = n;
	}
	while (*cmp != 0);
	ASSERT(raw_areas < MAX_AREAS);
	return;
}

local BOOL DecompressRegionAreaExit(float a, float b, float c, float* best_fraction)
{
	float fraction;

	// Work out if the ordinate has left the region
	if ((b < a) || (b > c))
		return FALSE;
	fraction = (b - a) / (c - a);
	ASSERT((fraction >= 0) && (fraction <= 1));
	if (fraction > *best_fraction)
		return FALSE;
	*best_fraction = fraction;
	return TRUE;
}

#define PointInside1(p,l,u)	(p > l) && (p < u)
#define PointInside(p,l,u)	(PointInside1((p).x, (l).x, (u).x) && PointInside1((p).y, (l).y, (u).y))

local void DecompressRegionAreaTransit(double distance, int transit, double* best_distance, int* best_transit)
{
	// Work out if this is a better transit entry point
	if ((distance > 0) && (distance < *best_distance))
	{
		*best_distance = distance;
		*best_transit = transit;
	}
	return;
}

local void DecompressPoint(const MAPPOINT* pt, const MAPPOINT* rmin)
{
	// Clamp a value between two bounds
	ASSERT(raw_points < MAX_POINTS);
	raw_point[raw_points].x = pt->x - rmin->x;
	ASSERT((raw_point[raw_points].x >= -0.000001f) && (raw_point[raw_points].x <= 15));
	raw_point[raw_points].y = pt->y - rmin->y;
	ASSERT((raw_point[raw_points].y >= -0.000001f) && (raw_point[raw_points].y <= 30));
	DEBUG2((" DecompressPoint(%d) = (%g,%g)\r\n", raw_points, (double)raw_point[raw_points].x, (double)raw_point[raw_points].y));
	raw_points++;
	return;
}

local void DecompressRegionAreaOutside(const MAPPOINT* rmin, const MAPPOINT* rmax, int area, int start)
{
	TRANSIT transit[16];
	int transits = 0;
	TRANSIT* tp = NULL;
	const MAPPOINT* pt = &area_point[raw_area[area].point_offset];
	int i, n = raw_area[area+1].point_offset - raw_area[area].point_offset;
	MAPPOINT prev = pt[start];
	int remaining;

	// The first point is outside, the next is inside
	ASSERT(n >= 3);
	ASSERT(!PointInside(pt[start], *rmin, *rmax));
	ASSERT(PointInside(pt[(start + 1) % n], *rmin, *rmax));
	for (i = 1; i <= n; i++)
	{
		MAPPOINT next = pt[(start + i) % n];
		if (tp == NULL)
		{
			// We were outside the region
			ASSERT(!PointInside(prev, *rmin, *rmax));
			if (PointInside(next, *rmin, *rmax))
			{
				// We've come back inside
				float fraction = 1.f;
				ASSERT(transits < dimof(transit));
				tp = &transit[transits];
				tp->in_edge = 0;
				if (DecompressRegionAreaExit(-next.y, -rmin->y, -prev.y, &fraction))
					tp->in_edge = 1;
				if (DecompressRegionAreaExit(next.x, rmax->x, prev.x, &fraction))
					tp->in_edge = 2;
				if (DecompressRegionAreaExit(next.y, rmax->y, prev.y, &fraction))
					tp->in_edge = 3;
				if (DecompressRegionAreaExit(-next.x, -rmin->x, -prev.x, &fraction))
					tp->in_edge = 4;
				ASSERT(tp->in_edge > 0);
				tp->in_point.x = next.x + fraction * (prev.x - next.x);
				tp->in_point.y = next.y + fraction * (prev.y - next.y);
				tp->in_offset = (start + i) % n;
			}
		}
		else
		{
			// We were inside the region
			float fraction = 1.f;
			tp->out_edge = 0;
			if (DecompressRegionAreaExit(-prev.y, -rmin->y, -next.y, &fraction))
				tp->out_edge = 1;
			if (DecompressRegionAreaExit(prev.x, rmax->x, next.x, &fraction))
				tp->out_edge = 2;
			if (DecompressRegionAreaExit(prev.y, rmax->y, next.y, &fraction))
				tp->out_edge = 3;
			if (DecompressRegionAreaExit(-prev.x, -rmin->x, -next.x, &fraction))
				tp->out_edge = 4;
			if (tp->out_edge > 0)
			{
				// Elvis has left the building
				tp->out_point.x = prev.x + fraction * (next.x - prev.x);
				tp->out_point.y = prev.y + fraction * (next.y - prev.y);
				tp->out_offset = (start + i) % n;
				transits++;
				ASSERT(transits < dimof(transit));
				tp = NULL;
			}
		}
		prev = next;
	}
	ASSERT(transits > 0);
	ASSERT(tp == NULL);

	// Iterate around the transits joining the sections in order
	remaining = transits;
	do
	{
		// Look for an unused entry point
		int original_transit, next_transit, first = raw_points;
		raw_instruction[raw_instructions++] = area;
		for (original_transit = transits - 1; original_transit > 0; original_transit--)
			if (transit[original_transit].in_edge > 0)
				break;
		ASSERT(transit[original_transit].in_edge > 0);
		next_transit = original_transit;
		do
		{
			MAPPOINT next;
			int edge;

			// Join the entry to the exit
			ASSERT(remaining > 0);
			tp = &transit[next_transit];
			DecompressPoint(&tp->in_point, rmin);
			i = tp->in_offset;
			while (i != tp->out_offset)
			{
				ASSERT(i < n);
				DecompressPoint(&pt[i], rmin);
				if (++i == n)
					i = 0;
				ASSERT(i != tp->in_offset);
			}
			next = tp->out_point;
			edge = tp->out_edge;
			ASSERT(edge > 0);

			// Go clockwise around the edge until we get to another entry point
			next_transit = -1;
			do
			{
				double distance;
				ASSERT((raw_points - first) < 1000);
				ASSERT(edge > 0);
				switch (edge)
				{
					case 1:
						// Top edge
						next.y = rmin->y;
						DecompressPoint(&next, rmin);
						distance = rmax->x - next.x;
						ASSERT(distance > 0);
						for (i = 0; i < transits; i++)
							if (transit[i].in_edge == 1)
								DecompressRegionAreaTransit(transit[i].in_point.x - next.x, i, &distance, &next_transit);
						while (distance > 1)
						{
							next.x += 1;
							distance -= 1;
							DecompressPoint(&next, rmin);
						}
						if (next_transit < 0)
							edge++;
						break;

					case 2:
						// Right edge
						next.x = rmax->x;
						DecompressPoint(&next, rmin);
						distance = rmax->y - next.y;
						ASSERT(distance > 0);
						for (i = 0; i < transits; i++)
							if (transit[i].in_edge == 2)
								DecompressRegionAreaTransit(transit[i].in_point.y - next.y, i, &distance, &next_transit);
						while (distance > 1)
						{
							next.y += 1;
							distance -= 1;
							DecompressPoint(&next, rmin);
						}
						if (next_transit < 0)
							edge++;
						break;

					case 3:
						// Bottom edge
						next.y = rmax->y;
						DecompressPoint(&next, rmin);
						distance = next.x - rmin->x;
						ASSERT(distance > 0);
						for (i = 0; i < transits; i++)
							if (transit[i].in_edge == 3)
								DecompressRegionAreaTransit(next.x - transit[i].in_point.x, i, &distance, &next_transit);
						while (distance > 1)
						{
							next.x -= 1;
							distance -= 1;
							DecompressPoint(&next, rmin);
						}
						if (next_transit < 0)
							edge++;
						break;

					case 4:
						// Left edge
						next.x = rmin->x;
						DecompressPoint(&next, rmin);
						distance = next.y - rmin->y;
						ASSERT(distance > 0);
						for (i = 0; i < transits; i++)
							if (transit[i].in_edge == 4)
								DecompressRegionAreaTransit(next.y - transit[i].in_point.y, i, &distance, &next_transit);
						while (distance > 1)
						{
							next.y -= 1;
							distance -= 1;
							DecompressPoint(&next, rmin);
						}
						if (next_transit < 0)
							edge = 1;
						break;

					default:
						// Not here!
						UNREACHABLE();
				}
			}
			while (next_transit < 0);
			transit[next_transit].in_edge = 0;
			remaining--;
		}
		while (next_transit != original_transit);

		// This is the end of this polygon
		i = raw_points - first;
		ASSERT(i >= 3);
		raw_instruction[raw_instructions++] = i;
	}
	while (remaining > 0);
	return;
}

local void DecompressRegionAreaInside(const MAPPOINT* rmin, const MAPPOINT* rmax, int area)
{
	const MAPPOINT* pt = &area_point[raw_area[area].point_offset];
	int i, n = raw_area[area+1].point_offset - raw_area[area].point_offset;

	// The first point is inside
	raw_instruction[raw_instructions++] = area;

	// Work out the polygons for this area in this region
	ASSERT(n >= 3);
	for (i = 0; i < n; i++)
	{
		ASSERT(PointInside(pt[i], *rmin, *rmax));
		DecompressPoint(&pt[i], rmin);
	}
	raw_instruction[raw_instructions++] = n;
	ASSERT(raw_instructions < dimof(raw_instruction));
	return;
}

local void DecompressRegionArea(const MAPPOINT* rmin, const MAPPOINT* rmax, int area, int start)
{
	const MAPPOINT* pt = &area_point[raw_area[area].point_offset];
	int i, n = raw_area[area+1].point_offset - raw_area[area].point_offset;

	// Look for a point outside the region
	ASSERT(n >= 3);
	ASSERT(PointInside(pt[start], *rmin, *rmax));
	if (start > 0)
	{
		// The point immediately before must be outside
		DEBUG2(("DecompressRegionAreaOutside(%d)\r\n", area));
		DecompressRegionAreaOutside(rmin, rmax, area, start - 1);
		return;
	}
	for (i = n - 1; i > 0; i--)
		if (!PointInside(pt[i], *rmin, *rmax))
		{
			// Found the last point outside the region
			DEBUG2(("DecompressRegionAreaOutside(%d)\r\n", area));
			DecompressRegionAreaOutside(rmin, rmax, area, i);
			return;
		}
	DEBUG2(("DecompressRegionAreaInside(%d)\r\n", area));
	DecompressRegionAreaInside(rmin, rmax, area);
	return;
}

local void DecompressRegionOcean(void)
{
	MAPPOINT pt = { 0, 0 };

	// Draw a rectangle for the ocean in each area
	raw_instruction[raw_instructions++] = 0;
	for (; pt.x < 15; pt.x += 1)
		raw_point[raw_points++] = pt;
	for (pt.y = 0; pt.y < 30; pt.y += 1)
		raw_point[raw_points++] = pt;
	for (pt.x = 15; pt.x > 0; pt.x -= 1)
		raw_point[raw_points++] = pt;
	for (pt.y = 30; pt.y >= 0; pt.y -= 1)
		raw_point[raw_points++] = pt;
	raw_instruction[raw_instructions++] = 91;
	ASSERT((pt.x == 0) && (pt.y == -1));
	return;
}

local void DecompressRegion(const MAPPOINT* rmin, const MAPPOINT* rmax)
{
	int a, p;

	// Decompress all the areas that must be drawn in a rectangular map region
	ASSERT(rmin->x < rmax->x);
	ASSERT(rmin->y < rmax->y);
	DecompressRegionOcean();
	for (a = 0; a < raw_areas; a++)
		for (p = raw_area[a].point_offset; p < raw_area[a+1].point_offset; p++)
			if (PointInside(area_point[p], *rmin, *rmax))
			{
				DecompressRegionArea(rmin, rmax, a, p - raw_area[a].point_offset);
				break;
			}
	return;
}

void Decompress(void)
{
	MAPPOINT rmin, rmax;
	int x, y;

	// Decompress the coastline data
	DecompressCountryNames();
	raw_instructions = 0;
	raw_points = 0;
	raw_instructions = 0;
	raw_points = 0;
	DecompressAreas();
	rmin.y = 0;
	rmax.y = 30;
	for (y = 0; y < 6; y++)
	{
		rmin.x = -180;
		rmax.x = -165;
		for (x = 0; x < 24; x++)
		{
			DecompressRegion(&rmin, &rmax);
			ASSERT(raw_instructions < dimof(raw_instruction));
			raw_instruction[raw_instructions++] = -1;
			rmin.x = rmax.x;
			rmax.x += 15;
		}
		rmin.y = rmax.y;
		rmax.y += 30;
	}
	return;
}
