//----------------------------------------------------------------------------
//	SOURCE:	MapConv.c
//	AUTHOR:	Ian Taylor
//	NOTES:	Conversion utility
//----------------------------------------------------------------------------

#include "Map.h"
#include <stdio.h>
#include <stdlib.h>

// Intrinsic mathematical functions
#pragma intrinsic(fabs, sqrt, exp, pow, log, sin, cos, tan, atan, atan2)

#define DONE_EMR(emr) OutputDebugString("EMR" #emr "\r\n")
//#define DONE_EMR(emr) fprintf(fp, "EMR" #emr "\n")
#define DO_EMR(emr) case EMR_##emr: DONE_EMR(emr); break;

// To text
local const RECTL reset = { 0, 0, 1000, 1000 };
local const XFORM xreset = { 1, 0, 0, 1, 0, 0 };
local COLORREF colour[256];
local COLORREF brush_colour;
local RECTL window;
local RECTL viewport;
local XFORM transform;
local BOOL dirty;

// Robinson projection
local void FromRobinson(double* x, double* y);

local int CALLBACK ConvertToTextEnum(HDC hdc, LPHANDLETABLE table, const ENHMETARECORD* rec, int ntable, LPARAM param)
{
	FILE* fp = (FILE*)param;
	int i;

	// Enumerates the EMF records in the clipboard
	USE(hdc);
	USE(table);
	USE(ntable);
	ASSERT(fp != NULL);
	switch (rec->iType)
	{
		case EMR_HEADER:
			// Reset the colours
			for (i = 0; i < dimof(colour); i++)
				colour[i] = 0xDEADBEEF;
			brush_colour = 0x00808080;
			window = reset;
			viewport = reset;
			transform = xreset;
			dirty = TRUE;
			DONE_EMR(HEADER);
			break;

		case EMR_CREATEBRUSHINDIRECT:
			{
				const EMRCREATEBRUSHINDIRECT* brush = (EMRCREATEBRUSHINDIRECT*)rec;
				ASSERT(brush->ihBrush < dimof(colour));
				colour[brush->ihBrush] = brush->lb.lbColor;
			}
			DONE_EMR(CREATEBRUSHINDIRECT);
			break;

		case EMR_SELECTOBJECT:
			{
				const EMRSELECTOBJECT* select = (EMRSELECTOBJECT*)rec;
				if (select->ihObject < dimof(colour))
					if (colour[select->ihObject] != 0xDEADBEEF)
						brush_colour = colour[select->ihObject];
			}
			DONE_EMR(SELECTOBJECT);
			break;

		case EMR_POLYGON16:
			{
				const EMRPOLYGON16* polygon16 = (EMRPOLYGON16*)rec;
				if (dirty)
				{
					fprintf(fp, "WINDOW,%d,%d,%d,%d\n", window.left, window.top, window.right, window.bottom);
					fprintf(fp, "VIEWPORT,%d,%d,%d,%d\n", viewport.left, viewport.top, viewport.right, viewport.bottom);
					fprintf(fp, "TRANSFORM,%g,%g,%g,%g,%g,%g\n", transform.eM11, transform.eM12, transform.eM21, transform.eM22, transform.eDx, transform.eDy);
					dirty = FALSE;
				}
				ASSERT(polygon16->apts[0].x == polygon16->apts[polygon16->cpts - 1].x);
				ASSERT(polygon16->apts[0].y == polygon16->apts[polygon16->cpts - 1].y);
				fprintf(fp, "POLYGON,%d,0x%06X", polygon16->cpts, brush_colour);
				for (i = 0; i < (int)polygon16->cpts; i++)
					fprintf(fp, ", %d,%d", (int)polygon16->apts[i].x, (int)polygon16->apts[i].y);
				fprintf(fp, "\n");
			}
			break;

		case EMR_SETWINDOWORGEX:
			{
				const EMRSETWINDOWORGEX* windoworg = (EMRSETWINDOWORGEX*)rec;
				window.left = windoworg->ptlOrigin.x;
				window.top = windoworg->ptlOrigin.y;
				dirty = TRUE;
			}
			DONE_EMR(SETWINDOWORGEX);
			break;

		case EMR_SETWINDOWEXTEX:
			{
				const EMRSETWINDOWEXTEX* windowext = (EMRSETWINDOWEXTEX*)rec;
				window.right = windowext->szlExtent.cx;
				window.bottom = windowext->szlExtent.cy;
				dirty = TRUE;
			}
			DONE_EMR(SETWINDOWEXTEX);
			break;

		case EMR_SETVIEWPORTORGEX:
			{
				const EMRSETVIEWPORTORGEX* viewportorg = (EMRSETVIEWPORTORGEX*)rec;
				viewport.left = viewportorg->ptlOrigin.x;
				viewport.top = viewportorg->ptlOrigin.y;
				dirty = TRUE;
			}
			DONE_EMR(SETVIEWPORTORGEX);
			break;

		case EMR_SETVIEWPORTEXTEX:
			{
				const EMRSETVIEWPORTEXTEX* viewportext = (EMRSETVIEWPORTEXTEX*)rec;
				viewport.right = viewportext->szlExtent.cx;
				viewport.bottom = viewportext->szlExtent.cy;
				dirty = TRUE;
			}
			DONE_EMR(SETVIEWPORTEXTEX);
			break;

		case EMR_SETWORLDTRANSFORM:
			transform = ((EMRSETWORLDTRANSFORM*)rec)->xform;
			dirty = TRUE;
			DONE_EMR(SETWORLDTRANSFORM);
			break;

		case EMR_CREATEPEN:
		case EMR_DELETEOBJECT:
		case EMR_CREATEMONOBRUSH:
		case EMR_SETBKMODE:
			// Ignore
			break;

		case EMR_SETMAPMODE:
			// OOPS
			ASSERT(rec->dParm[0] == MM_ANISOTROPIC);
			DONE_EMR(SETMAPMODE);
			break;

		DO_EMR(POLYBEZIER)
		DO_EMR(POLYGON)
		DO_EMR(POLYLINE)
		DO_EMR(POLYBEZIERTO)
		DO_EMR(POLYLINETO)
		DO_EMR(POLYPOLYLINE)
		DO_EMR(POLYPOLYGON)
//		DO_EMR(SETWINDOWEXTEX)
//		DO_EMR(SETWINDOWORGEX)
//		DO_EMR(SETVIEWPORTEXTEX)
//		DO_EMR(SETVIEWPORTORGEX)
		DO_EMR(SETBRUSHORGEX)
		DO_EMR(EOF)
		DO_EMR(SETPIXELV)
		DO_EMR(SETMAPPERFLAGS)
//		DO_EMR(SETMAPMODE)
//		DO_EMR(SETBKMODE)
		DO_EMR(SETPOLYFILLMODE)
		DO_EMR(SETROP2)
		DO_EMR(SETSTRETCHBLTMODE)
		DO_EMR(SETTEXTALIGN)
		DO_EMR(SETCOLORADJUSTMENT)
		DO_EMR(SETTEXTCOLOR)
		DO_EMR(SETBKCOLOR)
		DO_EMR(OFFSETCLIPRGN)
		DO_EMR(MOVETOEX)
		DO_EMR(SETMETARGN)
		DO_EMR(EXCLUDECLIPRECT)
		DO_EMR(INTERSECTCLIPRECT)
		DO_EMR(SCALEVIEWPORTEXTEX)
		DO_EMR(SCALEWINDOWEXTEX)
		DO_EMR(SAVEDC)
		DO_EMR(RESTOREDC)
//		DO_EMR(SETWORLDTRANSFORM)
		DO_EMR(MODIFYWORLDTRANSFORM)
//		DO_EMR(SELECTOBJECT)
//		DO_EMR(CREATEPEN)
//		DO_EMR(CREATEBRUSHINDIRECT)
//		DO_EMR(DELETEOBJECT)
		DO_EMR(ANGLEARC)
		DO_EMR(ELLIPSE)
		DO_EMR(RECTANGLE)
		DO_EMR(ROUNDRECT)
		DO_EMR(ARC)
		DO_EMR(CHORD)
		DO_EMR(PIE)
		DO_EMR(SELECTPALETTE)
		DO_EMR(CREATEPALETTE)
		DO_EMR(SETPALETTEENTRIES)
		DO_EMR(RESIZEPALETTE)
		DO_EMR(REALIZEPALETTE)
		DO_EMR(EXTFLOODFILL)
		DO_EMR(LINETO)
		DO_EMR(ARCTO)
		DO_EMR(POLYDRAW)
		DO_EMR(SETARCDIRECTION)
		DO_EMR(SETMITERLIMIT)
		DO_EMR(BEGINPATH)
		DO_EMR(ENDPATH)
		DO_EMR(CLOSEFIGURE)
		DO_EMR(FILLPATH)
		DO_EMR(STROKEANDFILLPATH)
		DO_EMR(STROKEPATH)
		DO_EMR(FLATTENPATH)
		DO_EMR(WIDENPATH)
		DO_EMR(SELECTCLIPPATH)
		DO_EMR(ABORTPATH)
		DO_EMR(GDICOMMENT)
		DO_EMR(FILLRGN)
		DO_EMR(FRAMERGN)
		DO_EMR(INVERTRGN)
		DO_EMR(PAINTRGN)
		DO_EMR(EXTSELECTCLIPRGN)
		DO_EMR(BITBLT)
		DO_EMR(STRETCHBLT)
		DO_EMR(MASKBLT)
		DO_EMR(PLGBLT)
		DO_EMR(SETDIBITSTODEVICE)
		DO_EMR(STRETCHDIBITS)
		DO_EMR(EXTCREATEFONTINDIRECTW)
		DO_EMR(EXTTEXTOUTA)
		DO_EMR(EXTTEXTOUTW)
		DO_EMR(POLYBEZIER16)
//		DO_EMR(POLYGON16)
		DO_EMR(POLYLINE16)
		DO_EMR(POLYBEZIERTO16)
		DO_EMR(POLYLINETO16)
		DO_EMR(POLYPOLYLINE16)
		DO_EMR(POLYPOLYGON16)
		DO_EMR(POLYDRAW16)
//		DO_EMR(CREATEMONOBRUSH)
		DO_EMR(CREATEDIBPATTERNBRUSHPT)
		DO_EMR(EXTCREATEPEN)
		DO_EMR(POLYTEXTOUTA)
		DO_EMR(POLYTEXTOUTW)
		DO_EMR(SETICMMODE)
		DO_EMR(CREATECOLORSPACE)
		DO_EMR(SETCOLORSPACE)
		DO_EMR(DELETECOLORSPACE)
		DO_EMR(GLSRECORD)
		DO_EMR(GLSBOUNDEDRECORD)
		DO_EMR(PIXELFORMAT)

		default:
			BREAKPOINT("Unknown EMR code");
	}
	return TRUE;
}

void ConvertToText(void)
{
	HENHMETAFILE hclip;

	// Convert EMF to text
	VERIFY(OpenClipboard(NULL));
	hclip = (HENHMETAFILE)GetClipboardData(CF_ENHMETAFILE);
	if (hclip == NULL)
		MessageBeep(0);
	else
	{
		DWORD err;
		FILE* fp = fopen("map_text.txt", "w");
		VERIFY(EnumEnhMetaFile(NULL, hclip, ConvertToTextEnum, fp, NULL));
		err = GetLastError();
		fclose(fp);
	}
	VERIFY(CloseClipboard());
	return;
}

local int WriteVector(FILE* fp_out, int angle, int length)
{
	// Write a single vector entry
	if (length == 0)
		return -1;
	if (length < 0)
	{
		length = -length;
		angle ^= 0x80;
	}
	while (length > 255)
	{
		fprintf(fp_out, ", %d,%d", angle, 255);
		length -= 255;
	}
	fprintf(fp_out, ", %d,%d", angle, length);
	return angle;
}

#define ASSERT_POINT(x, y)	ASSERT((x > -180.01) && (x < 180.01) && (y > -90.01) && (y < 90.01))

HENHMETAFILE ConvertToEMFFile(FILE* fp_in, FILE* fp_out, BOOL synopsis)
{
	HDC hdc;
	int n, i, x, y;
	BOOL error = FALSE;
	int cycle = 0x00FF00;
	char namebuf[256];

	// Create a EMF from text
	brush_colour = 0xDEADBEEF;
	window = reset;
	viewport = reset;
	transform = xreset;
	hdc = CreateEnhMetaFile(NULL, "map.emf", NULL, "Map Text Conversion");
	ASSERT(hdc != NULL);
	VERIFY(SetMapMode(hdc, MM_ANISOTROPIC));
	if (!synopsis)
		SelectObject(hdc, GetStockObject(NULL_PEN));
	while (!error && !feof(fp_in))
	{
		i = fgetc(fp_in);
		if (i == '#')
		{
			while (fgetc(fp_in) != '\n')
				;
			fscanf(fp_in, "\n");
			continue;
		}
		ungetc(i, fp_in);
		if (fscanf(fp_in, "POLYGON,%d,0x%X,\"%[^\"]\"", &n, &i, namebuf) == 3)
		{
			int triangle = (namebuf[0] == '*');
			int reverse = (namebuf[triangle] == '~');
			const char* name= &namebuf[triangle + reverse];
			double max_d = 0;
			POINT pt[1024];
			struct FPOINT { double x, y; } fpt[1024];
			ASSERT(n <= dimof(pt));
			ASSERT(n <= dimof(fpt));
			if (i == 0xFFFFFF)
				i = cycle++;
			if ((COLORREF)i != brush_colour)
			{
				HBRUSH hbrush;
				brush_colour = (COLORREF)i;
				hbrush = SelectObject(hdc, CreateSolidBrush(brush_colour));
				ASSERT(hbrush != NULL);
				VERIFY(DeleteObject(hbrush));
			}
			if (!synopsis)
				fprintf(fp_out, "POLYGON,%d,0x%06lX,\"%s\"", n-1, i, name);
			for (i = 0; !error && (i < n); i++)
				if (fscanf(fp_in, ", %d,%d", &x, &y) == 2)
				{
					// OOPS
					fpt[i].x = (x - 0) / 32350.0 * 360 - 180;
					fpt[i].y = (y - 231) / 16408.0 * 180 - 90;
					FromRobinson(&fpt[i].x, &fpt[i].y);
					if (i > 0)
					{
						double d = hypot(fpt[i].x - fpt[i - 1].x, fpt[i].y - fpt[i - 1].y);
						if (d > max_d)
							max_d = d;
					}
				}
				else
					error = TRUE;
			if (!error && reverse)
			{
				// Reverse the points
				for (i = 0; i < (n / 2); i++)
				{
					struct FPOINT swap = fpt[i];
					fpt[i] = fpt[n-i-1];
					fpt[n-i-1] = swap;
				}
			}
			if (!error)
			{
				double lx = 0, ly = 0;
				int la = -1, ll = 0, pa = -2;
				for (i = 0; i < n; i++)
				{
					double fx = fpt[i].x;
					double fy = fpt[i].y;
					ASSERT_POINT(fx, fy);
					if (!synopsis)
						fprintf(fp_out, ", %.3f,%.3f", (double)fx, (double)fy);
					else if (i == 0)
					{
						// First point
						lx = fx;
						ly = fy;
						fprintf(fp_out, "AREA(%s, %.8f,%.8f, %.8f)", name, (double)lx, (double)ly, (double)max_d);
					}
					else
					{
						// Vector to next point
						double scale = max_d / 250;
						double dx = fx - lx;
						double dy = fy - ly;
						double dxy = hypot(dx, dy);
						int length = (int)floor(dxy / scale + 0.5);
						int angle = (int)floor(atan2(dx, dy) / M_PI * 128 + 0.5);
						if (angle < 0)
							angle += 256;
						ASSERT((length >= 0) && (length <= 255));
						ASSERT((angle >= 0) && (angle <= 256));
						if (length > 0)
						{
							// This is a non-trivial vector
							ASSERT(length <= 255);
							angle &= 0xFF;
							if (angle == la)
							{
								// Continue on at this angle
								ll += length;
							}
							else if (angle == (la ^ 0x80))
							{
								// Reverse up
								ll -= length;
							}
							else
							{
								// Write this vector
								int wa = WriteVector(fp_out, la, ll);
								if (wa >= 0)
								{
									ASSERT(pa != (wa ^ 0x80));
									pa = wa;
								}
								la = angle;
								ll = length;
							}
							lx += sin(angle * M_PI / 128) * length * scale;
							ly += cos(angle * M_PI / 128) * length * scale;
							ASSERT_POINT(lx, ly);
							ASSERT(hypot(fx - lx, fy - ly) < dxy);
						}
						else
						{
							// Skip this vector
							fx = lx;
							fy = ly;
						}
					}
					pt[i].x = (int)((fx + 180) * window.right / 360 + 0.5);
					pt[i].y = (int)((fy + 90) * window.bottom / 180 + 0.5);
				}
				fscanf(fp_in, "\n");
				VERIFY(Polygon(hdc, pt, n));
				if (synopsis)
				{
					int wa = WriteVector(fp_out, la, ll);
					if (wa >= 0)
					{
						ASSERT(pa != (wa ^ 0x80));
						pa = wa;
					}
					fprintf(fp_out, ", %d,\n", pa ^ 0x80);
				}
				else
					fprintf(fp_out, "\n");
			}
		}
		else if (fscanf(fp_in, "WINDOW,0,0,%d,%d\n", &x, &y) == 2)
		{
			window.right = x;
			window.bottom = y;
			VERIFY(SetWindowExtEx(hdc, x, y, NULL));
		}
		else if (fscanf(fp_in, "VIEWPORT,0,0,%d,%d\n", &x, &y) == 2)
		{
			viewport.right = x;
			viewport.bottom = y;
			VERIFY(SetViewportExtEx(hdc, x, y, NULL));
		}
		else if (fscanf(fp_in, "TRANSFORM,%g,0,0,%g,0,0\n", &transform.eM11, &transform.eM22) == 2)
		{
			VERIFY(SetGraphicsMode(hdc, GM_ADVANCED));
			VERIFY(SetWorldTransform(hdc, &transform));
		}
		else
		{
			fgets(namebuf, dimof(namebuf), fp_in);
			error = TRUE;
		}
	}
	if (brush_colour != 0xDEADBEEF)
		VERIFY(DeleteObject(SelectObject(hdc, GetStockObject(NULL_BRUSH))));
	if (error)
	{
		CloseEnhMetaFile(hdc);
		UNREACHABLE();
		return NULL;
	}
	return CloseEnhMetaFile(hdc);
}

local void ConvertToSomething(BOOL synopsis)
{
	HENHMETAFILE hclip = NULL;
	FILE* fp_in;
	FILE* fp_out;

	// Convert text to EMF/synopsis
	fp_in = fopen("map.txt", "r");
	if (fp_in != NULL)
	{
		fp_out = fopen(synopsis ? "map_syn.txt" : "map_out.txt", "w");
		if (fp_out != NULL)
		{
			hclip = ConvertToEMFFile(fp_in, fp_out, synopsis);
			fclose(fp_out);
		}
		fclose(fp_in);
	}
	if (hclip == NULL)
		MessageBeep(0);
	else
	{
		VERIFY(OpenClipboard(NULL));
		VERIFY(EmptyClipboard());
		VERIFY(SetClipboardData(CF_ENHMETAFILE, hclip));
		VERIFY(CloseClipboard());
		VERIFY(DeleteEnhMetaFile(hclip));
	}
	return;
}

void ConvertToEMF(void)
{
	// Convert text to EMF
	ConvertToSomething(FALSE);
	return;
}

void ConvertToSynopsis(void)
{
	// Convert text to EMF
	ConvertToSomething(TRUE);
	return;
}

local double LatitudeToRobinson(double lat)
{
	// Robinson (p.214)
	static const float cy[][4] =
	{
		{ 0.0000F,	 0.01240000F,	 3.72529e-10F,	 1.15484e-09F },
		{ 0.0620F,	 0.01240010F,	 1.76951e-08F,	-5.92321e-09F },
		{ 0.1240F,	 0.01239980F,	-7.09668e-08F,	 2.25753e-08F },
		{ 0.1860F,	 0.01240080F,	 2.66917e-07F,	-8.44523e-08F },
		{ 0.2480F,	 0.01239710F,	-9.99682e-07F,	 3.15569e-07F },
		{ 0.3100F,	 0.01241080F,	 3.73349e-06F,	-1.17790e-06F },
		{ 0.3720F,	 0.01235980F,	-1.39350e-05F,	 4.39588e-06F },
		{ 0.4340F,	 0.01255010F,	 5.20034e-05F,	-1.00051e-05F },
		{ 0.4968F,	 0.01231980F,	-9.80735e-05F,	 9.22397e-06F },
		{ 0.5571F,	 0.01203080F,	 4.02857e-05F,	-5.29010e-06F },
		{ 0.6176F,	 0.01203690F,	-3.90662e-05F,	 7.36117e-07F },
		{ 0.6769F,	 0.01170150F,	-2.80246e-05F,	-8.54283e-07F },
		{ 0.7346F,	 0.01135720F,	-4.08389e-05F,	-5.18524e-07F },
		{ 0.7903F,	 0.01090990F,	-4.86169e-05F,	-1.07180e-06F },
		{ 0.8435F,	 0.01034330F,	-6.46934e-05F,	 5.36384e-09F },
		{ 0.8936F,	 0.00969679F,	-6.46129e-05F,	-8.54894e-06F },
		{ 0.9394F,	 0.00840949F,	-0.000192847F,	-4.21023e-06F },
		{ 0.9761F,	 0.00616525F,	-0.000256001F,	-4.21021e-06F },
		{ 1.0000F,	 0.00000000F,	 0.000000000F,	 0.000000000F }
	};
	const int dx = (lat > 0) ? 90 : -90;
	double a = fabs(lat);
	const int j = (int)(a / 5);
	ASSERT((j >= 0) && (j < dimof(cy)));
	a -= j * 5;
	return dx * (cy[j][0] + a * (cy[j][1] + a * (cy[j][2] + a * cy[j][3])));
}

local double LongitudeToRobinson(double lon, double lat)
{
	// Robinson (p.214)
	static const float cx[][4] =
	{
		{ 1.0000F,	-5.67239e-12F,	-7.15511e-05F,	 3.11028e-06F },
		{ 0.9986F,	-0.000482241F,	-2.48970e-05F,	-1.33094e-06F },
		{ 0.9954F,	-0.000831031F,	-4.48610e-05F,	-9.86588e-07F },
		{ 0.9900F,	-0.00135363F,	-5.96598e-05F,	 3.67749e-06F },
		{ 0.9822F,	-0.00167442F,	-4.49750e-06F,	-5.72394e-06F },
		{ 0.9730F,	-0.00214869F,	-9.03565e-05F,	 1.88767e-08F },
		{ 0.9600F,	-0.00305084F,	-9.00732e-05F,	 1.64869e-06F },
		{ 0.9427F,	-0.00382792F,	-6.53428e-05F,	-2.61493e-06F },
		{ 0.9216F,	-0.00467747F,	-0.000104566F,	 4.81220e-06F },
		{ 0.8962F,	-0.00536222F,	-3.23834e-05F,	-5.43445e-06F },
		{ 0.8679F,	-0.00609364F,	-0.000113900F,	 3.32521e-06F },
		{ 0.8350F,	-0.00698325F,	-6.40219e-05F,	 9.34582e-07F },
		{ 0.7986F,	-0.00755337F,	-5.00038e-05F,	 9.35532e-07F },
		{ 0.7597F,	-0.00798325F,	-3.59716e-05F,	-2.27604e-06F },
		{ 0.7186F,	-0.00851366F,	-7.01120e-05F,	-8.63072e-06F },
		{ 0.6732F,	-0.00986209F,	-0.000199572F,	 1.91978e-05F },
		{ 0.6213F,	-0.01041800F,	 8.83948e-05F,	 6.24031e-06F },
		{ 0.5722F,	-0.00906601F,	 0.000181999F,	 6.24033e-06F },
		{ 0.5322F,	 0.00000000F,	 0.000000000F,	 0.000000000F }
	};
	double a = fabs(lat);
	const int j = (int)(a / 5);
	ASSERT((j >= 0) && (j < dimof(cx)));
	a -= j * 5;
	return lon * (cx[j][0] + a * (cx[j][1] + a * (cx[j][2] + a * cx[j][3])));
}

local void FromRobinson(double* x, double* y)
{
	// Find longitude and latitude from Robinson coords
	double lat = 0;
	double dlat = 45;
	double lon = 0;
	double dlon = 91;
	ASSERT((x != NULL) && (*x >= -180) && (*x <= 180));
	ASSERT((y != NULL) && (*y >= -90) && (*y <= 90));
	while (dlat > 0)
	{
		double ry = LatitudeToRobinson(lat);
		if (ry < *y)
			lat += dlat;
		else
			lat -= dlat;
		dlat /= 2;
	}
	ASSERT((lat >= -90) && (lat <= 90));
	while (dlon > 0)
	{
		double rx = LongitudeToRobinson(lon, lat);
		if (rx < *x)
			lon += dlon;
		else
			lon -= dlon;
		dlon /= 2;
	}
	*x = lon;
	*y = lat;
	ASSERT((lon > -181) && (lon < 181));
	return;
}

struct CPOINT
{
	int ix, iy;
	double fx, fy;
	int join[6];
	int joins;
	int maxjoins;
	int seg[6];
	int segs;
	int finalid;
};
struct CPOLY
{
	int v[1024];
	int vs;
	BOOL triangle;
	char name[32];
};
struct CSEGMENT
{
	int v[1024];
	int vs;
	int finalid;
	int used;
};
local struct CSTRUCT
{
	struct CPOINT pt[16384];
	int pts;
	struct CPOLY poly[512];
	int polys;
	struct CSEGMENT seg[1024];
	int segs;
	int pointid;
	int segmentid;
}
c;

local int AddCPoint(int x, int y)
{
	// Add a shared point
	int i;
	for (i = 0; i < c.pts; i++)
		if ((x == c.pt[i].ix) && (y == c.pt[i].iy))
			return i;
	ASSERT(i < dimof(c.pt));
	c.pt[i].ix = x;
	c.pt[i].iy = y;
	c.pt[i].fx = (x - 0) / 32350.0 * 360 - 180;
	c.pt[i].fy = (y - 231) / 16408.0 * 180 - 90;
	FromRobinson(&c.pt[i].fx, &c.pt[i].fy);
	c.pts++;
	return i;
}

local void SetCPointSegment(int v, int s)
{
	// Inform a point it's used in a segment
	int i;
	for (i = 0; i < c.pt[v].segs; i++)
		ASSERT(c.pt[v].seg[i] != s);
	ASSERT(c.pt[v].segs < dimof(c.pt[v].seg));
	c.pt[v].seg[c.pt[v].segs++] = s;
	return;
}

local void AddCJoin(int v1, int v2)
{
	// Add a join between two vertices
	int i;
	ASSERT(v1 != v2);
	for (i = 0; i < c.pt[v1].joins; i++)
		if (c.pt[v1].join[i] == v2)
			return;
	ASSERT(c.pt[v1].joins < dimof(c.pt[v1].join));
	c.pt[v1].join[c.pt[v1].joins++] = v2;
	c.pt[v1].maxjoins = c.pt[v1].joins;
	return;
}

local void RemoveCJoin(int v1, int v2)
{
	// Remove a join between two vertices
	int i, j = c.pt[v1].joins - 1;
	ASSERT(v1 != v2);
	ASSERT(j >= 0);
	for (i = j; i >= 0; i--)
		if (c.pt[v1].join[i] == v2)
			break;
	ASSERT(i >= 0);
	c.pt[v1].join[i] = c.pt[v1].join[j];
	c.pt[v1].join[j] = v2;
	c.pt[v1].joins--;
	return;
}

local void AddCSegment(int v1)
{
	// Add a segment between junctions
	int s = c.segs++;
	struct CSEGMENT* seg = &c.seg[s];
	int v2;
	ASSERT(c.segs <= dimof(c.seg));
	SetCPointSegment(v1, s);
	seg->v[0] = v1;
	seg->vs = 1;
	v2 = c.pt[v1].join[0];
	ASSERT(v1 != v2);
	RemoveCJoin(v1, v2);
	RemoveCJoin(v2, v1);
	while (c.pt[v2].maxjoins == 2)
	{
		// Step along the non-junctioned path
		ASSERT(c.pt[v2].joins == 1);
		ASSERT(seg->vs < dimof(seg->v));
		SetCPointSegment(v2, s);
		seg->v[seg->vs++] = v2;
		v1 = v2;
		v2 = c.pt[v2].join[0];
		ASSERT(v1 != v2);
		RemoveCJoin(v1, v2);
		RemoveCJoin(v2, v1);
		if (v2 == seg->v[0])
			break;
	}
	if (v2 != seg->v[0])
	{
		// This is not a loop
		SetCPointSegment(v2, s);
		seg->v[seg->vs++] = v2;
		seg->finalid = -1;
	}
	return;
}

local int FindCNeighbour(const struct CSEGMENT* seg, int i1, int dir)
{
	// Find the point adjacent to another (handle looping)
	int i2 = i1 + dir;
	ASSERT((i1 >= 0) || (i1 < seg->vs));
	ASSERT((dir == 1) || (dir == -1));
	if (i2 < 0)
		return (seg->finalid == 0) ? seg->v[seg->vs - 1] : -1;
	if (i2 >= seg->vs)
		return (seg->finalid == 0) ? seg->v[0] : -1;
	return seg->v[i2];
}

local struct CSEGMENT* FindCSegment(int v1, int v2, int* dir)
{
	// Find the segment with the joined pair of vertices
	struct CSEGMENT* seg = NULL;
	int i;
	c.pt[v1].joins = c.pt[v1].maxjoins;
	for (i = 0; i < c.pt[v1].joins; i++)
		if (c.pt[v1].join[i] == v2)
		{
			// We've found the join, find the common segment
			int x, y, z;
			for (x = 0; x < c.pt[v1].segs; x++)
				for (y = 0; y < c.pt[v2].segs; y++)
					if (c.pt[v1].seg[x] == c.pt[v2].seg[y])
					{
						// We've found a common segment
						struct CSEGMENT* ss = NULL;
						int found = 0;
						ss = &c.seg[c.pt[v1].seg[x]];
						for (z = 0; z < ss->vs; z++)
							if (ss->v[z] == v1)
							{
								// We've found the first vertex within the segment
								found++;
								if (FindCNeighbour(ss, z, 1) == v2)
								{
									ASSERT(seg == NULL);
									seg = ss;
									*dir = 1;
								}
								else if (FindCNeighbour(ss, z, -1) == v2)
								{
									ASSERT(seg == NULL);
									seg = ss;
									*dir = -1;
								}
							}
						ASSERT(found == 1);
					}
		}
	ASSERT(seg != NULL);
	ASSERT(*dir != 0);
	return seg;
}

local void FindCPolyNext(const struct CPOLY* poly, int* v1, int* v2)
{
	// Look for the next two points in the poly
	int i;
	for (i = 0; i < (poly->vs - 2); i++)
		if (poly->v[i] == *v1)
		{
			// Found it
			ASSERT(poly->v[i+1] == *v2);
			*v1 = *v2;
			*v2 = poly->v[i+2];
			return;
		}
	if (poly->v[poly->vs - 2] == *v1)
	{
		ASSERT(poly->v[poly->vs - 1] == *v2);
		*v1 = *v2;
		*v2 = poly->v[0];
		return;
	}
	if (poly->v[poly->vs - 1] == *v1)
	{
		ASSERT(poly->v[0] == *v2);
		*v1 = *v2;
		*v2 = poly->v[1];
		return;
	}
	UNREACHABLE();
	return;
}

local void WriteCPoints(FILE* fp, int v1, int v2, BOOL first)
{
	// Write out a specific pair of points
	BOOL c1 = (c.pt[v1].finalid == 0);
	BOOL c2 = (c.pt[v2].finalid == 0);
	if (c1)
		c.pt[v1].finalid = ++c.pointid;
	if (c2)
		c.pt[v2].finalid = ++c.pointid;
	if (first)
		fprintf(fp, "POINTS_%d%d(%d, %d)", c1, c2, c.pt[v1].finalid, c.pt[v2].finalid);
	else
	{
		ASSERT(c1 == FALSE);
		fprintf(fp, "POINTS__%d(%d, %d)", c2, c.pt[v1].finalid, c.pt[v2].finalid);
	}
	if (c1)
		fprintf(fp, " POINT(%.4f,%.4f)", c.pt[v1].fx, c.pt[v1].fy);
	if (c2)
		fprintf(fp, " POINT(%.4f,%.4f)", c.pt[v2].fx, c.pt[v2].fy);
	return;
}

local int __cdecl WriteCSegmentCompare(const void* a, const void* b)
{
	// Double ordering
	double va = *(const double*)a;
	double vb = *(const double*)b;
	return (va < vb) ? -1 : (va > vb) ? 1 : 0;
}

local int WriteCJoins(FILE* fp, const int* v, int vs, double d, int min_joins)
{
	// Write out the joins
	BYTE byte[1024];
	int bytes = 0;
	double lx = c.pt[v[0]].fx;
	double ly = c.pt[v[0]].fy;
	int i;
	ASSERT(vs < dimof(byte));
	for (i = 1; i < vs; i++)
	{
		double nx = c.pt[v[i]].fx;
		double ny = c.pt[v[i]].fy;
		double dist = hypot(nx - lx, ny - ly);
		while (dist > d)
		{
			// Head towards this point
			int ia = (int)floor(atan2(nx - lx, ny - ly) / M_PI * 128 + 0.5);
			if (ia < 0)
				ia += 256;
			ASSERT(bytes < dimof(byte));
			byte[bytes++] = (BYTE)ia;
			lx += sin(ia * M_PI / 128) * d;
			ly += cos(ia * M_PI / 128) * d;
			dist = hypot(nx - lx, ny - ly);
		}
	}

	// Remove redundant tranversals
	i = 0;
	while (i < (bytes - 1))
		if (byte[i+1] == (byte[i] ^ 0x80))
		{
			bytes -= 2;
			memcpy(&byte[i], &byte[i+2], bytes - i);
			i = 0;
		}
		else
			i++;
	ASSERT(bytes > 0);

	// Create a triangle if there aren't enough points
	if (bytes < min_joins)
	{
		byte[bytes] = (BYTE)(byte[bytes - 1] + 0xFF * 2 / 3);
		bytes++;
	}

	// Write the traversals
	fprintf(fp, "\n\t\t");
	for (i = 0; i < bytes; i++)
		fprintf(fp, "%d,", byte[i]);
	fprintf(fp, "%d,", byte[bytes - 1] ^ 0x80);
	return bytes;
}

local void WriteCSegmentEx(FILE* fp, const int* v, int vs, int min_joins, BOOL loop, BOOL triangle)
{
	// Write a segment of linked points
	double min_d = 999999999.0;
	double max_d = 0.0;
	double med_d;
	double sum_d = 0.0;
	double d[512];
	int ds = vs - 1;
	int js = 0;
	int i;
	if (loop)
	{
		// Close the loop
		ASSERT(ds < dimof(d));
		d[ds] = hypot(c.pt[v[0]].fx - c.pt[v[ds]].fx, c.pt[v[0]].fy - c.pt[v[ds]].fy);
		min_d = d[ds];
		max_d = d[ds];
		ds++;
	}
	ASSERT(ds <= dimof(d));
	for (i = 0; i < (vs - 1); i++)
	{
		d[i] = hypot(c.pt[v[i]].fx - c.pt[v[i+1]].fx, c.pt[v[i]].fy - c.pt[v[i+1]].fy);
		ASSERT(d[i] > 0);
		if (d[i] < min_d)
			min_d = d[i];
		if (d[i] > max_d)
			max_d = d[i];
		sum_d += d[i];
	}

	// Find the median length
	qsort(d, ds, sizeof(d[0]), WriteCSegmentCompare);
	med_d = d[(ds - 1) / 2];
	if (triangle)
	{
		// Special case equilateral triangle
		double cx = 0.0;
		double cy = 0.0;
		double furthest = 0.0;
		double angle = 0;
		int n = vs - 1;
		for (i = 0; i < n; i++)
		{
			// Find the centroid
			cx += c.pt[v[i]].fx;
			cy += c.pt[v[i]].fy;
		}
		cx /= n;
		cy /= n;
		for (i = 0; i < n; i++)
		{
			double dx = c.pt[v[i]].fx - cx;
			double dy = c.pt[v[i]].fy - cy;
			double distance = dx * dx + dy * dy;
			if (distance > furthest)
			{
				furthest = distance;
				angle = atan2(dy, dx);
			}
		}
		angle = (angle / M_PI + 1) * 180;
		fprintf(fp, "\tTRIANGLE(%.4f,%.4f) LENGTH(%.2f) TWIST(%.2f)", cx, cy, sum_d * 0.19245, angle);
	}
	else if (vs > 2)
	{
		// Non-trivial join
		double dd;
		int vv = (vs < 55) ? (vs + 45) : 100;
		i = (int)floor(min_d * (100 - vv) + med_d * vv + 0.001);
		if (i == 0)
			i = 1;
		else if (i > 255)
			i = 255;
		ASSERT((i > 0) && (i < 256));
		dd = i * 0.01;
		if (vs <= (min_joins + 1))
		{
			// Bad triangle
			dd = (dd + min_d) / 2;
		}
		fprintf(fp, " LENGTH(%.2f)", dd);
		js = WriteCJoins(fp, v, vs, dd, min_joins);
		ASSERT(js >= min_joins);
	}
	fprintf(fp, "\n\t\t// MIN=%g MAX=%g MEDIAN=%g SUM=%g POINTS=%d JOINS=%d\n", min_d, max_d, med_d, sum_d, vs, js);
	return;
}

local void ReverseCSegment(int v[], int n)
{
	// Reverse the elements in a segment
	int i;
	for (i = 0; i < (n / 2); i++)
	{
		int swap = v[i];
		v[i] = v[n-i-1];
		v[n-i-1] = swap;
	}
	return;
}

local int WriteCSegment(FILE* fp, struct CSEGMENT* seg, int dir, int* vn1, int* vn2, BOOL first, BOOL triangle)
{
	// Write a loop or segment section
	ASSERT(seg != NULL);
	ASSERT((dir == 1) || (dir == -1));
	if (fp != NULL)
	{
		switch (seg->finalid)
		{
			case 0:
				// This is a loop
				ASSERT(seg->used == 1);
				if (!triangle)
					fprintf(fp, "\tLOOP(%.4f,%.4f)", c.pt[seg->v[0]].fx, c.pt[seg->v[0]].fy);
				if (dir < 0)
					ReverseCSegment(seg->v + 1, seg->vs - 1);
				WriteCSegmentEx(fp, &seg->v[0], seg->vs, 2, TRUE, triangle);
				return -1;

			case -1:
				// This is a new segment
				ASSERT(!triangle);
				ASSERT(seg->used >= 1);
				if (seg->used > 1)
				{
					seg->finalid = ++c.segmentid;
					if (seg->vs > 2)
						fprintf(fp, "\tSEGMENT_(%d) ", seg->finalid);
					else
						fprintf(fp, "\tLINE_(%d) ", seg->finalid);
				}
				else
				{
					seg->finalid = 999999;
					if (seg->vs > 2)
						fprintf(fp, "\tSEGMENT__ ");
					else
						fprintf(fp, "\tLINE__ ");
				}
				if (dir < 0)
				{
					ReverseCSegment(seg->v, seg->vs);
					dir = 1;
				}
				WriteCPoints(fp, seg->v[0], seg->v[seg->vs - 1], first);
				WriteCSegmentEx(fp, &seg->v[0], seg->vs, 1, FALSE, FALSE);
				break;

			default:
				// A known segment
				ASSERT(!triangle);
				ASSERT(seg->used == 2);
				fprintf(fp, "\tSEGMENT(%d)\n", seg->finalid * -dir);
				break;
		}
	}
	else
	{
		// Just remember how much we're used
		seg->used++;
		if (seg->finalid == 0)
			return -1;
	}

	// Return the last two point in this non-loop segment
	if (dir > 0)
	{
		// Going forwards
		*vn1 = seg->v[seg->vs - 2];
		*vn2 = seg->v[seg->vs - 1];
		return seg->v[0];
	}

	// Going backwards
	*vn1 = seg->v[1];
	*vn2 = seg->v[0];
	return seg->v[seg->vs - 1];
}

local void WriteCPoly(FILE* fp, const struct CPOLY* poly)
{
	// Write out a final poly
	int dir, v0, vn1, vn2;
	struct CSEGMENT* seg = FindCSegment(poly->v[0], poly->v[1], &dir);
	if (fp != NULL)
		fprintf(fp, "AREA(%s)\n", poly->name);
	v0 = WriteCSegment(fp, seg, dir, &vn1, &vn2, TRUE, poly->triangle);
	if (v0 >= 0)
	{
		// That wasn't a loop
		ASSERT(seg->finalid != 0);
		FindCPolyNext(poly, &vn1, &vn2);
		ASSERT(vn1 != v0);
		do
		{
			// Continue back to the start point
			seg = FindCSegment(vn1, vn2, &dir);
			ASSERT(seg->finalid != 0);
			ASSERT(!poly->triangle);
			VERIFY(WriteCSegment(fp, seg, dir, &vn1, &vn2, FALSE, FALSE) >= 0);
			FindCPolyNext(poly, &vn1, &vn2);
		}
		while (vn1 != v0);
		if (fp != NULL)
			fprintf(fp, "\tSEGMENT_END\n");
	}
	return;
}

local BOOL ConvertToCFile(FILE* fp_in, FILE* fp_out)
{
	struct CPOLY* poly = c.poly;
	int n, i, x, y;
	char namebuf[256];

	// Create a final C file from text
	memset(&c, 0, sizeof(c));
	window = reset;
	transform = xreset;
	while (!feof(fp_in))
	{
		i = fgetc(fp_in);
		if (i == '#')
		{
			while (fgetc(fp_in) != '\n')
				;
			fscanf(fp_in, "\n");
			continue;
		}
		ungetc(i, fp_in);
		if (fscanf(fp_in, "POLYGON,%d,0x%X,\"%[^\"]\"", &n, &i, namebuf) == 3)
		{
			int triangle = (namebuf[0] == '*');
			int reverse = (namebuf[triangle] == '~');
			ASSERT(c.polys < dimof(c.poly));
			strcpy(poly->name, &namebuf[reverse + triangle]);
			ASSERT(n <= dimof(poly->v));
			for (i = 0; i < n; i++)
				if (fscanf(fp_in, ", %d,%d", &x, &y) == 2)
					poly->v[i] = AddCPoint(x, y);
				else
					return FALSE;
			if (!reverse)
			{
				// Reverse the points
				for (i = 0; i < (n / 2); i++)
				{
					int swap = poly->v[i];
					poly->v[i] = poly->v[n-i-1];
					poly->v[n-i-1] = swap;
				}
			}
			poly->vs = --n;
			poly->triangle = triangle;

			// Join the dots
			ASSERT(poly->v[0] == poly->v[n]);
			for (i = 0; i < n; i++)
			{
				AddCJoin(poly->v[i], poly->v[i+1]);
				AddCJoin(poly->v[i+1], poly->v[i]);
			}
			fscanf(fp_in, "\n");
			c.polys++;
			poly++;
		}
		else if (fscanf(fp_in, "WINDOW,0,0,%d,%d\n", &x, &y) == 2)
		{
			window.right = x;
			window.bottom = y;
		}
		else if (fscanf(fp_in, "VIEWPORT,0,0,%d,%d\n", &x, &y) == 2)
		{
			// Skip
		}
		else if (fscanf(fp_in, "TRANSFORM,%g,0,0,%g,0,0\n", &transform.eM11, &transform.eM22) == 2)
		{
			// Skip
		}
		else if (fscanf(fp_in, "STOP\n") == 0)
		{
			// Stop
			break;
		}
		else
		{
			fgets(namebuf, dimof(namebuf), fp_in);
			return FALSE;
		}
	}

	// Look for junctions
	for (i = 0; i < c.pts; i++)
		while ((c.pt[i].joins > 0) && (c.pt[i].maxjoins > 2))
			AddCSegment(i);

	// Look for dead ends
	for (i = 0; i < c.pts; i++)
		if (c.pt[i].joins == 1)
			AddCSegment(i);

	// Look for simple loops (islands)
	for (i = 0; i < c.pts; i++)
	{
		while (c.pt[i].joins > 0)
			AddCSegment(i);
		ASSERT(c.pt[i].joins == 0);
	}

	// Restore all the joins
	for (i = 0; i < c.pts; i++)
		c.pt[i].joins = c.pt[i].maxjoins;

	// Measure all the polygons
	for (i = 0; i < c.polys; i++)
		WriteCPoly(NULL, &c.poly[i]);

	// Write out all the polygons
	for (i = 0; i < c.polys; i++)
		WriteCPoly(fp_out, &c.poly[i]);
	return TRUE;
}

void ConvertToFinal(void)
{
	BOOL okay = FALSE;
	FILE* fp_in;
	FILE* fp_out;

	// Convert text to final
	fp_in = fopen("map.txt", "r");
	if (fp_in != NULL)
	{
		fp_out = fopen("MapData.h", "w");
		if (fp_out != NULL)
		{
			okay = ConvertToCFile(fp_in, fp_out);
			fclose(fp_out);
		}
		fclose(fp_in);
	}
	if (!okay)
		MessageBeep(0);
	return;
}
