//----------------------------------------------------------------------------
//	SOURCE:	MapRep.c
//	AUTHOR:	Ian Taylor
//	NOTES:	Report window for Map2 demo application
//----------------------------------------------------------------------------

#include "Map.h"
#include "MapCodes.h"
#include "MapRep.h"
#include <float.h>

// Control IDs
#define IDC_COLOURS			100
#define IDC_MODE			101
#define IDC_NUMERATOR		102
#define IDC_DENOMINATOR		103
#define IDC_LISTVIEW		104

// The unique value for #N/A
local const double c_NA = DBL_MAX;

// The unpacked values
#define MAX_STATS	53
#define MAX_CODES	10
#define FIRST_CODE	3
#define FIRST_STAT	(FIRST_CODE + MAX_CODES)
#define MAX_COLUMNS	(FIRST_CODE + MAX_CODES + MAX_STATS)
typedef struct _MAPSTATS
{
	double statistic;
	TCHAR code[MAX_CODES][8];
	double number[MAX_STATS];
	int cc4;
	struct _STAT
	{
		double colour;	// [0,1]
		double value;	// Not logarithm
		int rank;		// 0 for N/A
		BOOL equal;		// Equal rank?
	}
	legend;
}
MAPSTATS;
local MAPSTATS raw_stats[Code_CountryCount+1];

// Information about each unit
typedef struct _MAPUNIT
{
	LPCTSTR name;
	int suffix;
}
MAPUNIT;
local MAPUNIT raw_unit[Unit_Count];

// Information about each statistic
typedef struct _MAPCOLUMN
{
	enum Unit unit;
	TCHAR name[64];
	int width;
}
MAPCOLUMN;
local MAPCOLUMN raw_column[MAX_COLUMNS];

// Current map colours
local COLORREF map_colours[Code_CountryCount+1][2];

// Suffixes
enum Suffix { Suffix_None = 0, Suffix_Decimal, Suffix_Percent };

local LPTSTR ReportFormatPositive(LPTSTR p, double number, int sig, BOOL dp)
{
	// Format a positive number
	double e = log10(number);
	int m, j, i = (int)e;
	TCHAR digit[16];
	int digits = 0;
	ASSERT(number > 0);
	if (e >= 0)
		i++;
	m = (int)(number / pow(10, i - sig) + 0.5);
	ASSERT(m > 0);
	while ((m % 10) == 0)
	{
		// Skip trailing zeroes
		m /= 10;
	}
	do
	{
		// Work out the significant digit characters in reverse order
		digit[digits++] = (TCHAR)(m % 10 + '0');
		m /= 10;
	}
	while (m > 0);
	if (i <= 0)
	{
		// Decimal point to the left of the digits
		for (j = 0; j < digits; j++)
			*--p = digit[j];
		while (i++ < 0)
			*--p = '0';
		*--p = '.';
		*--p = '0';
	}
	else if (i >= digits)
	{
		// No decimal points
		int thousands = -1;
		if (dp)
		{
			// Force a decimal point
			*--p = '0';
			*--p = '.';
		}
		while (i-- > digits)
		{
			if ((thousands++ > 0) && ((thousands % 3) == 0))
				*--p = ',';
			*--p = '0';
		}
		for (j = 0; j < digits; j++)
		{
			if ((thousands++ > 0) && ((thousands % 3) == 0))
				*--p = ',';
			*--p = digit[j];
		}
	}
	else
	{
		// Decimal point somewhere in the midst of the digits
		int thousands = -1;
		for (j = 0; j < digits; j++)
		{
			if ((thousands >= 0) && (thousands++ > 3) && ((thousands % 3) == 0))
				*--p = ',';
			*--p = digit[j];
			if (++i == digits)
			{
				thousands = 2;
				*--p = '.';
			}
		}
	}
	return p;
}

local LPTSTR ReportFormatNumber(LPTSTR p, double number, int sig, int suffix)
{
	// Format a number
	if (p == NULL)
	{
		// Use a local buffer
		static TCHAR buffer[32];
		p = buffer + dimof(buffer);
		*--p = '\0';
	}
	if (number != c_NA)
	{
		if (suffix & Suffix_Percent)
			*--p = '%';
		if (number == 0)
		{
			// Zero
			if (suffix & Suffix_Decimal)
			{
				*--p = '0';
				*--p = '.';
			}
			*--p = '0';
		}
		else if (number > 0)
		{
			// Positive
			p = ReportFormatPositive(p, number, sig, (suffix & Suffix_Decimal));
		}
		else
		{
			// Negative
			p = ReportFormatPositive(p, -number, sig, (suffix & Suffix_Decimal));
			*--p = '-';
		}
	}
	return p;
}

local LPTSTR ReportFormatOrdinal(LPTSTR p, int value)
{
	// Format an ordinal (e.g 1st, 2nd, etc)
	ASSERT(p != NULL);
	ASSERT(value > 0);
	if (((value % 10) == 1) && (value != 11))
	{
		*--p = 't';
		*--p = 's';
	}
	else if (((value % 10) == 2) && (value != 12))
	{
		*--p = 'd';
		*--p = 'n';
	}
	else if (((value % 10) == 3) && (value != 13))
	{
		*--p = 'd';
		*--p = 'r';
	}
	else
	{
		*--p = 'h';
		*--p = 't';
	}
	while (value > 0)
	{
		*--p = (TCHAR)((value % 10) + '0');
		value /= 10;
	}
	return p;
}

local LPTSTR ReportFormatStatistic(int country, BOOL rank)
{
	// Format the statistic
	LPTSTR p = NULL;
	if (rank && raw_stats[country].legend.rank)
	{
		// Format the rank
		static TCHAR buffer[32];
		p = buffer + dimof(buffer);
		*--p = '\0';
		*--p = ')';
		p = ReportFormatOrdinal(p, raw_stats[country].legend.rank);
		if (raw_stats[country].legend.equal)
			*--p = '=';
		*--p = '(';
		*--p = ' ';
	}
	return ReportFormatNumber(p, raw_stats[country].statistic, 6, Suffix_None);
}

local LPCTSTR ReportFormat(int row, int column)
{
	// Format a report cell
	int c = column - 3;
	if (c < -1)
		return region_names[row+1];
	if (c == -1)
		return ReportFormatStatistic(row+1, FALSE);
	if (c < MAX_CODES)
		return raw_stats[row+1].code[c];
	c -= MAX_CODES;
	ASSERT(c < MAX_STATS);
	return ReportFormatNumber(NULL, raw_stats[row+1].number[c], 6, raw_unit[raw_column[column].unit].suffix);
}

local void ReportUnpackChars(LPTSTR code, Bits a0, Bits a1, Bits a2, char offset)
{
	// Store up to three characters in a code buffer
	if (a0 != 0)
	{
		code[0] = (TCHAR)(a0 + offset - 1);
		if (a1 != 0)
		{
			code[1] = (TCHAR)(a1 + offset - 1);
			if (a2 != 0)
				code[2] = (TCHAR)(a2 + offset - 1);
		}
	}
	return;
}

local void ReportUnpackInteger(LPTSTR code, Bits v)
{
	// Store up to three digits in a code buffer
	ASSERT((v >= 0) && (v <= 999));
	if (v >= 100)
		*(code++) = (TCHAR)(v / 100 + '0');
	if (v >= 10)
		*(code++) = (TCHAR)((v / 10) % 10 + '0');
	if (v > 0)
		*(code++) = (TCHAR)(v % 10 + '0');
	return;
}

local double ReportUnpackFloat(Bits m, int s)
{
	// Unpack three significant digits
	if (m == 1023)
		return c_NA;
	ASSERT((m >= 0) && (m <= 1000));
	return m * pow(10, s);
}

local void ReportUnpack(void)
{
	LPCTSTR c = COLUMNS;
	int i;
	memset(raw_unit, 0, sizeof(raw_unit));
	memset(raw_column, 0, sizeof(raw_column));
	for (i = 0; i < dimof(raw_column); i++)
	{
		// Load column headings
		int unit = *(c++) - 'A';
		ASSERT((unit >= 0) && (unit < dimof(raw_unit)));
		raw_column[i].unit = (enum Unit)unit;
		if (raw_unit[unit].name == NULL)
		{
			// This is a new unit
			raw_unit[unit].name = c;
			switch (*c)
			{
				case '%':
					// Percentage
					raw_unit[unit].suffix = Suffix_Decimal | Suffix_Percent;
					break;
				case 'y':
					// Age in years
					raw_unit[unit].suffix = Suffix_Decimal;
					break;
			}
			while (*(c++) != '\0');
		}
		raw_column[i].width = *(c++) - '0';
		OS(lstrcpy)(raw_column[i].name, c);
		if (raw_unit[unit].name[0] != '\0')
		{
			OS(lstrcat)(raw_column[i].name, _T(" ("));
			OS(lstrcat)(raw_column[i].name, raw_unit[unit].name);
			OS(lstrcat)(raw_column[i].name, _T(")"));
		}
		while (*(c++) != '\0');
	}
	memset(raw_stats, 0, sizeof(raw_stats));
	for (i = 1; i < dimof(raw_stats); i++)
	{
		const MAPNUMPACKED* p = &raw_packed[i-1];
		MAPSTATS* s = &raw_stats[i];
		double* f = s->number - 1;
		s->statistic = c_NA;

		// Unpack the textual stats
		ReportUnpackChars(s->code[0], p->a11, p->a12, 0, 'A');
		ReportUnpackChars(s->code[1], p->a21, p->a22, p->a23, 'A');
		ReportUnpackChars(s->code[2], p->a31, p->a32, 0, 'a');
		s->code[3][0] = '.';
		ReportUnpackChars(&s->code[3][1], p->a41, p->a42, 0, 'a');
		ReportUnpackChars(s->code[4], p->a51, p->a52, p->a53, 'A');
		ReportUnpackChars(s->code[5], p->a61, p->a62, p->a63, 'A');
		ReportUnpackChars(s->code[6], p->a71, p->a72, p->a73, 'A');
		s->code[7][0] = (TCHAR)(p->vv ? 'Y' : 'N');
		ReportUnpackInteger(s->code[8], p->m01);
		if (p->m02 > 0)
		{
			s->code[9][0] = '+';
			if (p->m02 > 999)
			{
				// Prefix "1-" to the ITU calling
				s->code[9][1] = '1';
				s->code[9][2] = '-';
				ReportUnpackInteger(&s->code[9][3], p->m02 - 1000);
			}
			else
				ReportUnpackInteger(&s->code[9][1], p->m02);
		}

		// Unpack the numeric stats
		*++f = ReportUnpackFloat(p->m03, p->e03-2);
		*++f = ReportUnpackFloat(p->m04, p->e04-2);
		*++f = ReportUnpackFloat(p->m05, p->e05-2);
		ASSERT(f[-1] != c_NA);
		f[1] = (f[0] == c_NA) ? f[-1] : (f[-1] + f[0]);
		f++;
		*++f = ReportUnpackFloat(p->m06, p->e06);
		if (p->s06)
			*f = -*f;
		*++f = ReportUnpackFloat(p->m07, p->e07-2);
		*++f = ReportUnpackFloat(p->m08, p->e08-4);
		*++f = ReportUnpackFloat(p->m09, p->e09);
		*++f = ReportUnpackFloat(p->m10, p->e10);
		*++f = ReportUnpackFloat(p->m11, p->e11);
		*++f = ReportUnpackFloat(p->m12, -1);
		*++f = ReportUnpackFloat(p->m13, -1);
		*++f = ReportUnpackFloat(p->m14, p->e14);
		*++f = ReportUnpackFloat(p->m15, p->e15-1);
		*++f = ReportUnpackFloat(p->m16, p->e16-1);
		if (p->s16)
			*f = -*f;
		f[1] = (f[0] == c_NA) ? c_NA : f[-2] - f[-1] + f[0];
		f++;
		*++f = ReportUnpackFloat(p->m17, -2);
		if (*f != c_NA)
			*f += 0.79;
		*++f = ReportUnpackFloat(p->m18, p->e18-3);
		*++f = ReportUnpackFloat(p->m19, -2);
		*++f = ReportUnpackFloat(p->m20, -2);
		if (*f != c_NA)
			*f -= 1;
		*++f = ReportUnpackFloat(p->m21, -1);
		*++f = ReportUnpackFloat(p->m22, -1);
		*++f = ReportUnpackFloat(p->m23, -1);
		*++f = ReportUnpackFloat(p->m24, p->e24-2);
		*++f = ReportUnpackFloat(p->m25, p->e25-2);
		*++f = ReportUnpackFloat(p->m26, p->e26-3);
		if (p->s26)
			*f = -*f;
		*++f = ReportUnpackFloat(p->m27, p->e27);
		*++f = ReportUnpackFloat(p->m28, p->e28-3);
		*++f = ReportUnpackFloat(p->m29, p->e29-6);
		*++f = ReportUnpackFloat(p->m30, -3);
		*++f = ReportUnpackFloat(p->m31, p->e31-1);
		*++f = ReportUnpackFloat(p->m32, p->e32-1);
		f[1] = (f[0] == c_NA) ? c_NA : (f[0] - f[-1]);
		f++;
		*++f = ReportUnpackFloat(p->m33, -1);
		*++f = ReportUnpackFloat(p->m34, -1);
		*++f = ReportUnpackFloat(p->m35, p->e35-2);
		*++f = ReportUnpackFloat(p->m36, p->e36-3);
		*++f = ReportUnpackFloat(p->m37, p->e37-2);
		*++f = ReportUnpackFloat(p->m38, p->e38-1);
		*++f = ReportUnpackFloat(p->m39, p->e39-3);
		*++f = ReportUnpackFloat(p->m40, p->e40);
		*++f = ReportUnpackFloat(p->m41, p->e41-1);
		*++f = ReportUnpackFloat(p->m42, p->e42);
		*++f = ReportUnpackFloat(p->m43, p->e43);
		*++f = ReportUnpackFloat(p->m44, p->e44);
		*++f = ReportUnpackFloat(p->m45, p->e45-2);
		*++f = ReportUnpackFloat(p->m46, p->e46-2);
		*++f = ReportUnpackFloat(p->m47, p->e47-2);
		*++f = ReportUnpackFloat(p->m48, p->e48-2);
		*++f = ReportUnpackFloat(p->m49, -1);
		*++f = ReportUnpackFloat(p->m50, -1);
		*++f = ReportUnpackFloat(p->m51, p->e51-1);
//		*++f = ReportUnpackFloat(p->m52, p->e52-1);
		*++f = ReportUnpackFloat(p->m53, p->e53-2);
		ASSERT(f == &s->number[MAX_STATS-1]);

		// Unpack the four-colour value
		s->cc4 = p->cc;
	}
	return;
}

// Flag icon sizes
#define MAP_FLAG_ICON_SX	26
#define MAP_FLAG_ICON_SY	13

local HIMAGELIST ReportImageList(void)
{
	// Create an image list of flags
	const int sx = MAP_FLAG_ICON_SX * 32;
	const int sy = MAP_FLAG_ICON_SY * 32;
	HDC hdc1 = OS(CreateCompatibleDC)(NULL);
	HBITMAP hbitmap1 = OS(CreateCompatibleBitmap)(OS(GetDC)(NULL), sx, sy);
	HGDIOBJ hother1 = OS(SelectObject)(hdc1, hbitmap1);
	HDC hdc2 = OS(CreateCompatibleDC)(NULL);
	HBITMAP hbitmap2 = OS(CreateCompatibleBitmap)(OS(GetDC)(NULL), MAP_FLAG_ICON_SX, MAP_FLAG_ICON_SY);
	HIMAGELIST imagelist = OS(ImageList_Create)(MAP_FLAG_ICON_SX, MAP_FLAG_ICON_SY, ILC_COLOR24, Code_CountryCount, 0);
	int country;
	VERIFY(OS(SetStretchBltMode)(hdc2, STRETCH_HALFTONE));
	for (country = 0; country < Code_CountryCount; country++)
	{
		HGDIOBJ hother2 = OS(SelectObject)(hdc2, hbitmap2);
		FlagDraw(hdc1, sx, sy, 2.0, country+1);
		VERIFY(OS(StretchBlt)(hdc2, 0, 0, MAP_FLAG_ICON_SX, MAP_FLAG_ICON_SY, hdc1, 0, 0, sx, sy, SRCCOPY));
		VERIFY(OS(SelectObject)(hdc2, hother2));
		OS(ImageList_Add)(imagelist, hbitmap2, NULL);
	}
	VERIFY(OS(DeleteObject)(hbitmap2));
	VERIFY(OS(DeleteDC)(hdc2));
	VERIFY(OS(SelectObject)(hdc1, hother1));
	VERIFY(OS(DeleteObject)(hbitmap1));
	VERIFY(OS(DeleteDC)(hdc1));
	return imagelist;
}

local void ReportComboStrings(HWND hdlg, int item, LPCTSTR text)
{
	// Add a lines to a single combobox
	do
	{
		OS(SendDlgItemMessage)(hdlg, item, CB_ADDSTRING, 0, (LPARAM)text);
		while (*(text++));
	}
	while (*text);
	OS(SendDlgItemMessage)(hdlg, item, CB_SETCURSEL, 0, 0);
	return;
}

local void ReportFill(HWND hdlg)
{
	// Fill our report window
	static const TCHAR colours[] = _T("Political\0Spectrum\0Blue to Red\0Green to Magenta\0Brown to Blue\0Red Yellow Green\0Green Scale\0Grey Scale\0");
	static const TCHAR modes[] = _T("Standard\0Logarithmic\0Ranked\0");
	static const TCHAR none[] = _T("(None)");
	LVCOLUMN lvc = { LVCF_FMT|LVCF_TEXT|LVCF_WIDTH, LVCFMT_LEFT, 100 };
	int j, i = LVS_EX_FULLROWSELECT|LVS_EX_GRIDLINES|LVS_EX_HEADERDRAGDROP;
	int zero = 8;
	SIZE size;

	// Further initialisation
	ReportUnpack();
	OS(SendDlgItemMessage)(hdlg, IDC_LISTVIEW, LVM_SETEXTENDEDLISTVIEWSTYLE, i, i);
	OS(SendDlgItemMessage)(hdlg, IDC_LISTVIEW, LVM_SETIMAGELIST, LVSIL_SMALL, (LPARAM)ReportImageList());
	ReportComboStrings(hdlg, IDC_COLOURS, colours);
	ReportComboStrings(hdlg, IDC_MODE, modes);
	ReportComboStrings(hdlg, IDC_NUMERATOR, none);
	ReportComboStrings(hdlg, IDC_DENOMINATOR, none);

	// Find the width of a zero character
	if (OS(GetTextExtentPoint)(OS(GetDC)(OS(GetDlgItem)(hdlg, IDC_LISTVIEW)), _T("0000"), 4, &size))
		zero = (size.cx + 2) / 4;
	for (j = 0; j < MAX_COLUMNS; j++)
	{
		// Insert the column headers
		lvc.pszText = (LPTSTR)raw_column[j].name;
		lvc.cx = (raw_column[j].width + 4) * zero;
		if (raw_column[j].unit != Unit_Text)
		{
			lvc.fmt |= LVCFMT_RIGHT;
		}
		OS(SendDlgItemMessage)(hdlg, IDC_LISTVIEW, LVM_INSERTCOLUMN, j, (LPARAM)&lvc);
		if (j >= FIRST_STAT)
		{
			OS(SendDlgItemMessage)(hdlg, IDC_NUMERATOR, CB_ADDSTRING, 0, (LPARAM)lvc.pszText);
			OS(SendDlgItemMessage)(hdlg, IDC_DENOMINATOR, CB_ADDSTRING, 0, (LPARAM)lvc.pszText);
		}
	}
	for (i = 0; i < Code_CountryCount; i++)
	{
		// The first column is the flag
		LVITEM lvi = { LVIF_IMAGE|LVIF_PARAM };
		lvi.iItem = i;
		lvi.iImage = i;
		lvi.lParam = i+1;
		OS(SendDlgItemMessage)(hdlg, IDC_LISTVIEW, LVM_INSERTITEM, 0, (LPARAM)&lvi);
		lvi.mask = LVIF_TEXT;
		for (j = 1; j < MAX_COLUMNS; j++)
		{
			// Subsequent columns have content
			lvi.iSubItem = j;
			lvi.pszText = (LPTSTR)ReportFormat(i, j);
			OS(SendDlgItemMessage)(hdlg, IDC_LISTVIEW, LVM_SETITEM, 0, (LPARAM)&lvi);
		}
	}

	// Now fill the other controls
	return;
}

local int ReportSortCompareCode(LPCTSTR a, LPCTSTR b, int column, int ifequal)
{
	// Compare code
	int compare = OS(lstrcmpi)(a, b);
	if (compare == 0)
		return ifequal;
	if (*a == 0)
		return 1;
	if (*b == 0)
		return -1;
	return (column < 0) ? -compare : compare;
}

local int ReportSortCompareNumber(double a, double b, int column, int ifequal)
{
	// Compare code
	if (a == b)
		return ifequal;
	if (a == c_NA)
		return 1;
	if (b == c_NA)
		return -1;
	return (a < b) ? -column : column;
}

local int CALLBACK ReportSortCompare(LPARAM lparam1, LPARAM lparam2, LPARAM column)
{
	// Sort according to a column
	int c = (column < 0) ? (-FIRST_CODE - column) : (column - FIRST_CODE);
	if (c == -1)
		return ReportSortCompareNumber(raw_stats[lparam1].statistic, raw_stats[lparam2].statistic, column, lparam1 - lparam2);
	if (c < 0)
		return ReportSortCompareCode(region_names[lparam1], region_names[lparam2], column, lparam1 - lparam2);
	if (c < MAX_CODES)
		return ReportSortCompareCode(raw_stats[lparam1].code[c], raw_stats[lparam2].code[c], column, lparam1 - lparam2);
	c -= MAX_CODES;
	ASSERT(c < MAX_STATS);
	return ReportSortCompareNumber(raw_stats[lparam1].number[c], raw_stats[lparam2].number[c], column, lparam1 - lparam2);
}

local COLORREF ReportHSVtoRGB(double h, double s, double v)
{
	// Convert an HSV triplet to RGB
	BYTE l = (BYTE)(v * 255);
	ASSERT((h >= 0) && (h < 1));
	ASSERT((s >= 0) && (s <= 1));
	ASSERT((v >= 0) && (v <= 1));
	if (s != 0)
	{
		int i = (int)(h * 6);
		double f = (i & 1) ? (h * 6 - i) : (1 + i - h * 6);
		BYTE m = (BYTE)(v * 255 * (1 - s));
		BYTE n = (BYTE)(v * 255 * (1 - s * f));
		ASSERT((f >= 0) && (f <= 1));
		switch (i)
		{
			case 1:	return PALETTERGB(n, l, m);
			case 2:	return PALETTERGB(m, l, n);
			case 3:	return PALETTERGB(m, n, l);
			case 4:	return PALETTERGB(n, m, l);
			case 5:	return PALETTERGB(l, m, n);
		}
		return PALETTERGB(l, n, m);
	}
	return PALETTERGB(l, l, l);
} 

local void MapFillStatistics(HWND hdlg, int numerator, int denominator, int mode)
{
	// Fill in the statistics data for each country
	double rank[Code_CountryCount];
	int ranks = 0;
	int i;
	for (i = 1; i <= Code_CountryCount; i++)
	{
		// Sort the values
		MAPSTATS* s = &raw_stats[i];
		double value = (numerator >= 0) ? s->number[numerator] : c_NA;
		s->statistic = value;
		s->legend.value = value;
		s->legend.rank = (value != c_NA);
		if (!s->legend.rank)
			continue;

		// Look at the denominator too
		if (denominator >= 0)
		{
			double under = s->number[denominator];
			if ((under == c_NA) || (fabs(under) < DBL_EPSILON))
			{
				s->statistic = c_NA;
				s->legend.value = c_NA;
				s->legend.rank = 0;
				continue;
			}
			value /= under;
			s->statistic = value;
			s->legend.value = value;
		}

		// This is a valid statistic (not N/A)
		if ((ranks == 0) || (value < rank[ranks - 1]))
			rank[ranks++] = value;
		else
		{
			// Binary chop until we find the right place
			int lo = 0;
			int hi = ranks - 1;
			while (lo < hi)
			{
				int mid = (lo + hi) / 2;
				if (value < rank[mid])
					lo = mid + 1;
				else
					hi = mid;
			}
			ASSERT(lo == hi);
			ASSERT(value >= rank[lo]);
			for (hi = ranks; hi > lo; hi--)
				rank[hi] = rank[hi - 1];
			rank[hi] = value;
			ranks++;
		}
	}
	if (ranks > 0)
	{
		double lowest = rank[ranks - 1];
		double range = rank[0] - lowest;
		if (range <= 0.0)
			mode = -1;
		for (i = 1; i <= Code_CountryCount; i++)
		{
			MAPSTATS* s = &raw_stats[i];
			if (s->legend.rank)
			{
				// Rank the values
				double value = s->statistic;
				int lo = 0;
				int hi = ranks - 1;
				while (lo < hi)
				{
					int mid = (lo + hi) / 2;
					if (value < rank[mid])
						lo = mid + 1;
					else
						hi = mid;
				}
				ASSERT(lo == hi);
				ASSERT(value == rank[lo]);
				s->legend.rank = lo + 1;
				s->legend.equal = (lo < (ranks - 1)) && (rank[lo] == rank[lo+1]);
				switch (mode)
				{
					case 0:
						// Standard
						s->legend.colour = (value - lowest) / range;
						break;

					case 1:
						// Logarithm
						s->legend.colour = log10((value - lowest) / range * 9 + 1);
						break;

					case 2:
						// Ranked
						s->legend.colour = (double)lo / (ranks - 1);
						break;

					case -1:
						// Only a single value throughout
						s->legend.colour = 1;
						break;

					default:
						UNREACHABLE();
						break;
				}
				ASSERT((s->legend.colour >= 0) && (s->legend.colour <= 1));
			}
		}
	}

	// Now fill in the grid statistic column
	for (i = 0; i < Code_CountryCount; i++)
	{
		LVITEM lvi = { LVIF_PARAM };
		lvi.iItem = i;
		lvi.iSubItem = FIRST_CODE - 1;
		OS(SendDlgItemMessage)(hdlg, IDC_LISTVIEW, LVM_GETITEM, 0, (LPARAM)&lvi);
		lvi.pszText = ReportFormatStatistic(lvi.lParam, FALSE);
		lvi.mask = LVIF_TEXT;
		OS(SendDlgItemMessage)(hdlg, IDC_LISTVIEW, LVM_SETITEM, 0, (LPARAM)&lvi);
	}
	return;
}

// Helper macros for declaring dialog structures
#define DLG_DECLARE_PADDING(name, size, extra)	BYTE name##_padding[(3 - (((size) - 1) & 3)) + (extra)]
#define DLG_DECLARE_TEXT(name, text)			WCHAR name[sizeof(text)]
#define DLG_DECLARE_ITEM(name, caption)			\
	struct _##name								\
	{											\
		DLGITEMTEMPLATE itemtemplate;			\
		WORD class0, class1;					\
		DLG_DECLARE_TEXT(title, caption);		\
	}											\
	name;										\
	DLG_DECLARE_PADDING(name, sizeof(struct _##name)+2, 2)
#define DLG_DECLARE_ITEM_EX(name, classid)		\
	struct _##name								\
	{											\
		DLGITEMTEMPLATE itemtemplate;			\
		DLG_DECLARE_TEXT(classname, classid);	\
		DLG_DECLARE_TEXT(title, "");		\
	}											\
	name;										\
	DLG_DECLARE_PADDING(name, sizeof(struct _##name)+2, 2)

// Helper macros for defining dialog structure contents
#define DLG_TEXT(text)							L##text
#define DLG_DEFINE_TEXT(text)					DLG_TEXT(text)
#define DLG_DEFINE_ITEM(id, style, x,y, cx,cy, classid, caption)	\
	{											\
		{ WS_VISIBLE|style, 0, x,y, cx,cy, id },\
		0xFFFF,	classid,						\
		DLG_DEFINE_TEXT(caption)				\
	}, { 0 }
#define DLG_DEFINE_ITEM_EX(id, style, x,y, cx,cy, classid)	\
	{											\
		{ WS_VISIBLE|style, WS_EX_CLIENTEDGE, x,y, cx,cy, id },\
		DLG_DEFINE_TEXT(classid),				\
		DLG_DEFINE_TEXT("")				\
	}, { 0 }

// The dialog template header
#define DLG_CLASSNAME	"MapDialog"
#define DLG_CAPTION		"Map - Statistics"
#define DLG_FONT		"MS Sans Serif"
typedef struct _MAPDLGHEADER
{
	DLGTEMPLATE dlgtemplate;
	WORD menu;
	DLG_DECLARE_TEXT(classname, DLG_CLASSNAME);
	DLG_DECLARE_TEXT(caption, DLG_CAPTION);
	WORD fontsize;
	DLG_DECLARE_TEXT(fontname, DLG_FONT);
}
MAPDLGHEADER;

// The dialog template
local const struct _MAPDLGTEMPLATE
{
	MAPDLGHEADER header;
	DLG_DECLARE_ITEM_EX(item0, WC_LISTVIEWA);
	DLG_DECLARE_ITEM(item1, "Colours:");
	DLG_DECLARE_ITEM(item2, "");
	DLG_DECLARE_ITEM(item3, "Statistic:");
	DLG_DECLARE_ITEM(item4, "");
	DLG_DECLARE_ITEM(item5, "Mode:");
	DLG_DECLARE_ITEM(item6, "");
	DLG_DECLARE_ITEM(item7, "Divisor:");
	DLG_DECLARE_ITEM(item8, "");
}
mapdialog =
{
	{
		{ DS_CENTER|DS_SETFONT|WS_OVERLAPPED|WS_CAPTION|WS_SYSMENU|WS_THICKFRAME|WS_MAXIMIZEBOX, 0, 9, 0, 0, 282, 282 },
		0,
		DLG_DEFINE_TEXT(DLG_CLASSNAME),
		DLG_DEFINE_TEXT(DLG_CAPTION),
		8,
		DLG_DEFINE_TEXT(DLG_FONT)
	},
	DLG_DEFINE_ITEM_EX(IDC_LISTVIEW, LVS_REPORT|LVS_AUTOARRANGE|LVS_SHOWSELALWAYS, 0,32,282,250, WC_LISTVIEWA),
	DLG_DEFINE_ITEM(1, SS_RIGHT, 4,4, 25,10, 0x0082, "Colours:"),
	DLG_DEFINE_ITEM(IDC_COLOURS, CBS_DROPDOWNLIST, 30,2, 80,128, 0x0085, ""),
	DLG_DEFINE_ITEM(1, SS_RIGHT, 114,4, 27,10, 0x0082, "Statistic:"),
	DLG_DEFINE_ITEM(IDC_NUMERATOR, CBS_DROPDOWNLIST|WS_VSCROLL, 142,2, 140,128, 0x0085, ""),
	DLG_DEFINE_ITEM(1, SS_RIGHT, 4,18, 25,10, 0x0082, "Mode:"),
	DLG_DEFINE_ITEM(IDC_MODE, CBS_DROPDOWNLIST, 30,16, 80,128, 0x0085, ""),
	DLG_DEFINE_ITEM(1, SS_RIGHT, 114,18, 27,10, 0x0082, "Divisor:"),
	DLG_DEFINE_ITEM(IDC_DENOMINATOR, CBS_DROPDOWNLIST|WS_VSCROLL, 142,16, 140,128, 0x0085, ""),
};

// Global
local struct _STATE
{
	HWND hdlg;
	int sorted;
	BYTE highlighted[248];
}
state;

local void ReportSort(int column)
{
	// Sort by the appropriate column
	if (column <= 0)
		return;
	if (column == state.sorted)
		state.sorted = -column;
	else
		state.sorted = column;
    OS(SendDlgItemMessage)(state.hdlg, IDC_LISTVIEW, LVM_SORTITEMS, state.sorted, (LPARAM)ReportSortCompare);
	return;
}

void ReportSelect(int country)
{
	// Select a particular country in the list view
	LVITEM item = { 0, 0, 0, 0, LVIS_SELECTED|LVIS_FOCUSED };
	OS(SendDlgItemMessage)(state.hdlg, IDC_LISTVIEW, LVM_SETITEMSTATE, (WPARAM)-1, (LPARAM)&item);
	if (country > 0)
	{
		// Select a single item
		LVFINDINFO find = { LVFI_PARAM };
		int index;
		find.lParam = country;
		index = OS(SendDlgItemMessage)(state.hdlg, IDC_LISTVIEW, LVM_FINDITEM, (WPARAM)-1, (LPARAM)&find);
		ASSERT(index >= 0);
		item.state = item.stateMask;
		OS(SendDlgItemMessage)(state.hdlg, IDC_LISTVIEW, LVM_SETITEMSTATE, index, (LPARAM)&item);
		OS(SendDlgItemMessage)(state.hdlg, IDC_LISTVIEW, LVM_ENSUREVISIBLE, index, 0);
	}
	else
	{
		// Clear the selection mark too
		OS(SendDlgItemMessage)(state.hdlg, IDC_LISTVIEW, LVM_SETSELECTIONMARK, (WPARAM)-1, 0);
	}
	return;
}

local BOOL CALLBACK ReportDialogNotify(UINT id, NMLISTVIEW* nm)
{
	// Something has happened at a control
	switch (id)
	{
		case IDC_LISTVIEW:
			ASSERT(nm->hdr.hwndFrom == OS(GetDlgItem)(state.hdlg, IDC_LISTVIEW));
			switch (nm->hdr.code)
			{
				case LVN_ITEMCHANGED:
					// Something happened
					if (nm->uChanged & LVIF_STATE)
					{
						// Selection or status?
						LRESULT selected = OS(SendMessage)(nm->hdr.hwndFrom, LVM_GETITEMSTATE, nm->iItem, LVIS_SELECTED);
						state.highlighted[nm->lParam] = (BYTE)(selected != 0);
						if (nm->uNewState & LVIS_FOCUSED)
							FlagSelect(nm->lParam);
					}
					MapRecolour();
					MapRepaint();
					return TRUE;

				case LVN_COLUMNCLICK:
					// Column header was clicked
					ReportSort(nm->iSubItem);
					return TRUE;

				case LVN_KEYDOWN:
					// Key press
					switch (((NMLVKEYDOWN*)nm)->wVKey)
					{
						case VK_ESCAPE:
							// Clear all active
							FlagSelect(0);
							ReportSelect(0);
							MapRecolour();
							MapRepaint();
							break;
						case VK_RETURN:
							if (OS(SendMessage)(nm->hdr.hwndFrom, LVM_GETSELECTEDCOUNT, 0, 0) == 1)
							{
								// Zoom on map
								LVITEM lvi = { LVIF_PARAM };
								lvi.iItem = OS(SendMessage)(nm->hdr.hwndFrom, LVM_GETSELECTIONMARK, 0, 0);
								ASSERT(lvi.iItem >= 0);
								OS(SendMessage)(nm->hdr.hwndFrom, LVM_GETITEM, 0, (LPARAM)&lvi);
								RegionZoom(lvi.lParam);
							}
							break;
					}
					break;

				case NM_DBLCLK:
					// Double click
					if (nm->iItem >= 0)
					{
						LVITEM lvi = { LVIF_PARAM };
						lvi.iItem = nm->iItem;
						OS(SendMessage)(nm->hdr.hwndFrom, LVM_GETITEM, 0, (LPARAM)&lvi);
						RegionZoom(lvi.lParam);
					}
					return 0;
			}
	}
	return FALSE;
}

local BOOL CALLBACK ReportDialogProc(HWND hdlg, UINT umsg, WPARAM wparam, LPARAM lparam)
{
	// Handle dialog messages
	switch (umsg)
	{
		case WM_NOTIFY:
			// Something has happened at a control
			return ReportDialogNotify(wparam, (NMLISTVIEW*)lparam);

		case WM_COMMAND:
			// Normal control changes
			switch (wparam)
			{
				case MAKEWPARAM(IDC_COLOURS, CBN_SELCHANGE):
				case MAKEWPARAM(IDC_MODE, CBN_SELCHANGE):
				case MAKEWPARAM(IDC_NUMERATOR, CBN_SELCHANGE):
				case MAKEWPARAM(IDC_DENOMINATOR, CBN_SELCHANGE):
					// Inform the world
					MapRecolour();
					MapRepaint();
					break;
			}
			break;

		case WM_GETMINMAXINFO:
			// Don't let the window get too small
			{
				LPMINMAXINFO mmi = (LPMINMAXINFO)lparam;
				ASSERT(mmi != NULL);
				mmi->ptMinTrackSize.x = 256;
				mmi->ptMinTrackSize.y = 256;
			}
			return TRUE;

		case WM_SIZE:
			// Resize
			if (wparam != SIZE_MINIMIZED)
			{
				HWND hwnd = OS(GetDlgItem)(hdlg, IDC_LISTVIEW);
				RECT client, listview;
				VERIFY(OS(GetClientRect)(hdlg, &client));
				VERIFY(OS(GetWindowRect)(hwnd, &listview));
				VERIFY(OS(MapWindowPoints)(NULL, hdlg, (LPPOINT)&listview, 2));
				OS(MoveWindow)(hwnd, 0, listview.top, client.right, client.bottom - listview.top, TRUE);
			}
			return TRUE;

		case WM_CLOSE:
			// Actually, just hide it
			VERIFY(OS(ShowWindow)(hdlg, SW_HIDE));
			return 0;
	}
	return FALSE;
}

HWND ReportCreate(HWND hparent, HICON hicon)
{
	// Register the window class
	RECT parentRect, childRect;
	LONG deskWidth, deskHeight;
	WNDCLASS wc;
	memset(&state, 0, sizeof(state));
	OS(InitCommonControls)();
	wc.style = CS_HREDRAW|CS_VREDRAW|CS_OWNDC;
	wc.lpfnWndProc = OS(DefDlgProc);
	wc.cbClsExtra = 0;
	wc.cbWndExtra = DLGWINDOWEXTRA;
	wc.hInstance = (HINSTANCE)OS(GetWindowLong)(hparent, GWL_HINSTANCE);
	wc.hIcon = hicon;
	wc.hCursor = OS(LoadCursor)(NULL, IDC_ARROW);
	wc.hbrBackground = NULL;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = _T(DLG_CLASSNAME);
	ASSERT(wc.hCursor != NULL);
	VERIFY(OS(RegisterClass)(&wc));

	// Create the report dialog
    state.hdlg = OS(CreateDialogIndirectParam)(NULL, &mapdialog.header.dlgtemplate, hparent, ReportDialogProc, 0);
	ASSERT(state.hdlg != NULL);
	VERIFY(OS(GetWindowRect)(hparent, &parentRect));
	VERIFY(OS(GetWindowRect)(state.hdlg, &childRect));
	deskWidth = ((childRect.right + childRect.left - 1) | 1) + 1;
	deskHeight = ((childRect.bottom + childRect.top - 1) | 1) + 1;
	VERIFY(OS(MoveWindow)(state.hdlg, parentRect.right, 0, deskWidth - parentRect.right, deskHeight, FALSE));
	ReportFill(state.hdlg);
	OS(ShowWindow)(state.hdlg, SW_NORMAL);
	return state.hdlg;
}

local BOOL ReportHighlighted(int country)
{
	// Work out if a region is highlighted
	ASSERT((country >= 1) && (country < dimof(state.highlighted)));
	return state.highlighted[country];
}

void MapRecolour(void)
{
	// Recolour the whole map
	int scheme = OS(SendDlgItemMessage)(state.hdlg, IDC_COLOURS, CB_GETCURSEL, 0, 0);
	int numerator = OS(SendDlgItemMessage)(state.hdlg, IDC_NUMERATOR, CB_GETCURSEL, 0, 0) - 1;
	int denominator = OS(SendDlgItemMessage)(state.hdlg, IDC_DENOMINATOR, CB_GETCURSEL, 0, 0) - 1;
	int mode = OS(SendDlgItemMessage)(state.hdlg, IDC_MODE, CB_GETCURSEL, 0, 0);
	int country;
	OS(EnableWindow)(OS(GetDlgItem)(state.hdlg, IDC_NUMERATOR), (scheme != 0));
	OS(EnableWindow)(OS(GetDlgItem)(state.hdlg, IDC_DENOMINATOR), (scheme != 0) && (numerator >= 0));
	OS(EnableWindow)(OS(GetDlgItem)(state.hdlg, IDC_MODE), (scheme != 0) && (numerator >= 0));

	// Fill in the colours for each country depending on the statistics
	MapFillStatistics(state.hdlg, numerator, denominator, mode);
	map_colours[0][FALSE] = PALETTERGB(128,192,192);
	if (scheme == 0)
	{
		// Fill in the political country colours
		for (country = 1; country <= Code_CountryCount; country++)
		{
			double h = (4 >> raw_stats[country].cc4) / 6.0;
			map_colours[country][FALSE] = ReportHSVtoRGB(h, 0.25, 1);
			map_colours[country][TRUE] = ReportHSVtoRGB(h, 1, 1);
		}
	}
	else
	{
		// Fill in the statistical colour
		for (country = 1; country <= Code_CountryCount; country++)
			if (raw_stats[country].legend.rank)
			{
				double h, s, v = raw_stats[country].legend.colour;
				switch (scheme)
				{
					case 1:
						// Spectrum
						h = v * 0.866667;
						map_colours[country][FALSE] = ReportHSVtoRGB(h, 0.5, 0.5);
						map_colours[country][TRUE] = ReportHSVtoRGB(h, 1, 1);
						break;

					case 2:
						// Blue to Red
						h = v * 0.666667;
						s = fabs(v * 2 - 1);
						map_colours[country][FALSE] = ReportHSVtoRGB(h, s, 0.5);
						map_colours[country][TRUE] = ReportHSVtoRGB(h, s, 1);
						break;

					case 3:
						// Green to Magenta
						h = (v < 0.5) ? 0.333333 : 0.866667;
						s = fabs(v * 2 - 1);
						map_colours[country][FALSE] = ReportHSVtoRGB(h, s, 0.5);
						map_colours[country][TRUE] = ReportHSVtoRGB(h, s, 1);
						break;

					case 4:
						// Brown to Blue
						h = (v < 0.5) ? 0.0 : 0.5;
						s = fabs(v * 2 - 1);
						v /= 4;
						map_colours[country][FALSE] = ReportHSVtoRGB(h, s, v + 0.25);
						map_colours[country][TRUE] = ReportHSVtoRGB(h, s, v + 0.75);
						break;

					case 5:
						// Red Yellow Green
						v /= 3;
						map_colours[country][FALSE] = ReportHSVtoRGB(v, 0.7, 0.7);
						map_colours[country][TRUE] = ReportHSVtoRGB(v, 1, 1);
						break;

					case 6:
						// Green scale
						map_colours[country][FALSE] = ReportHSVtoRGB(0.333333, v, 1);
						map_colours[country][TRUE] = ReportHSVtoRGB(0, v, 1);
						break;

					case 7:
						// Grey scale
						v = (v + 1) / 2;
						map_colours[country][FALSE] = ReportHSVtoRGB(0, 0, v);
						map_colours[country][TRUE] = ReportHSVtoRGB(0, 1, v);
						break;
				}
			}
			else
			{
				// This is N/A
				switch (scheme)
				{
					case 7:
						// Grey scale
						map_colours[country][FALSE] = PALETTERGB(64,32,32);
						map_colours[country][TRUE] = PALETTERGB(160,64,64);
						break;

					default:
						// All others
						map_colours[country][FALSE] = PALETTERGB(32,32,32);
						map_colours[country][TRUE] = PALETTERGB(128,64,64);
						break;
				}
			}
	}
	return;
}

LPTSTR ReportTooltip(int region)
{
	// Format a tooltip window
	static TCHAR tooltip[256];
	int country = RegionToCountry(region);
	LPTSTR statistic = ReportFormatStatistic(country, TRUE);
	if (*statistic != '\0')
	{
		*--statistic = '\n';
		*--statistic = '\r';
	}
	if (region == country)
	{
		OS(wsprintf)(tooltip, _T("%s%s"), region_names[region], statistic);
	}
	else
	{
		OS(wsprintf)(tooltip, _T("%s (%s)%s"), region_names[region], region_names[country], statistic);
	}
	return tooltip;
}

COLORREF ColourDefault(int area)
{
	// Use the cached map colours
	if (area == 0)
		return map_colours[0][FALSE];
	area = RegionToCountry(raw_area[area].code);
	return map_colours[area][ReportHighlighted(area)];
}

COLORREF ColourGrey(int area)
{
	// Draw areas with their region colours
	BYTE grey = (BYTE)((area == 0) ? 0 : raw_area[area].code);
	return PALETTERGB(grey, grey, grey);
}
